#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n)google/appengine/datastore/snapshot.proto\x12\x13storage_onestore_v3\":\n\x08Snapshot\x12\n\n\x02ts\x18\x01 \x02(\x03\"\"\n\x06Status\x12\x0c\n\x08INACTIVE\x10\x00\x12\n\n\x06\x41\x43TIVE\x10\x01\x42G\n\x1e\x63om.google.storage.onestore.v3B\x10OnestoreSnapshotZ\x13storage_onestore_v3')



_SNAPSHOT = DESCRIPTOR.message_types_by_name['Snapshot']
_SNAPSHOT_STATUS = _SNAPSHOT.enum_types_by_name['Status']
Snapshot = _reflection.GeneratedProtocolMessageType('Snapshot', (_message.Message,), {
  'DESCRIPTOR' : _SNAPSHOT,
  '__module__' : 'google.appengine.datastore.snapshot_pb2'

  })
_sym_db.RegisterMessage(Snapshot)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\036com.google.storage.onestore.v3B\020OnestoreSnapshotZ\023storage_onestore_v3'
  _SNAPSHOT._serialized_start=66
  _SNAPSHOT._serialized_end=124
  _SNAPSHOT_STATUS._serialized_start=90
  _SNAPSHOT_STATUS._serialized_end=124




#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
















"""Higher-level Query wrapper.

There are perhaps too many query APIs in the world.

The fundamental API here overloads the 6 comparisons operators to
represent filters on property values, and supports AND and OR
operations (implemented as functions -- Python's 'and' and 'or'
operators cannot be overloaded, and the '&' and '|' operators have a
priority that conflicts with the priority of comparison operators).
For example::

  class Employee(Model):
    name = StringProperty()
    age = IntegerProperty()
    rank = IntegerProperty()

    @classmethod
    def demographic(cls, min_age, max_age):
      return cls.query().filter(AND(cls.age >= min_age, cls.age <= max_age))

    @classmethod
    def ranked(cls, rank):
      return cls.query(cls.rank == rank).order(cls.age)

  for emp in Employee.seniors(42, 5):
    print emp.name, emp.age, emp.rank

The 'in' operator cannot be overloaded, but is supported through the
IN() method.  For example::

  Employee.query().filter(Employee.rank.IN([4, 5, 6]))

Sort orders are supported through the order() method; unary minus is
overloaded on the Property class to represent a descending order::

  Employee.query().order(Employee.name, -Employee.age)

Besides using AND() and OR(), filters can also be combined by
repeatedly calling .filter()::

  q1 = Employee.query()  # A query that returns all employees
  q2 = q1.filter(Employee.age >= 30)  # Only those over 30
  q3 = q2.filter(Employee.age < 40)  # Only those in their 30s

A further shortcut is calling .filter() with multiple arguments; this
implies AND()::

  q1 = Employee.query()  # A query that returns all employees
  q3 = q1.filter(Employee.age >= 30,
                 Employee.age < 40)  # Only those in their 30s

And finally you can also pass one or more filter expressions directly
to the .query() method::

  q3 = Employee.query(Employee.age >= 30,
                      Employee.age < 40)  # Only those in their 30s

Query objects are immutable, so these methods always return a new
Query object; the above calls to filter() do not affect q1.  (On the
other hand, operations that are effectively no-ops may return the
original Query object.)

Sort orders can also be combined this way, and .filter() and .order()
calls may be intermixed::

  q4 = q3.order(-Employee.age)
  q5 = q4.order(Employee.name)
  q6 = q5.filter(Employee.rank == 5)

Again, multiple .order() calls can be combined::

  q5 = q3.order(-Employee.age, Employee.name)

The simplest way to retrieve Query results is a for-loop::

  for emp in q3:
    print emp.name, emp.age

Some other methods to run a query and access its results::

  q.iter() # Return an iterator; same as iter(q) but more flexible
  q.map(callback) # Call the callback function for each query result
  q.fetch(N) # Return a list of the first N results
  q.get() # Return the first result
  q.count(N) # Return the number of results, with a maximum of N
  q.fetch_page(N, start_cursor=cursor) # Return (results, cursor, has_more)

All of the above methods take a standard set of additional query
options, either in the form of keyword arguments such as
keys_only=True, or as QueryOptions object passed with
options=QueryOptions(...).  The most important query options are:

- keys_only: bool, if set the results are keys instead of entities
- limit: int, limits the number of results returned
- offset: int, skips this many results first
- start_cursor: Cursor, start returning results after this position
- end_cursor: Cursor, stop returning results after this position
- batch_size: int, hint for the number of results returned per RPC
- prefetch_size: int, hint for the number of results in the first RPC
- produce_cursors: bool, return Cursor objects with the results

For additional (obscure) query options and more details on them,
including an explanation of Cursors, see datastore_query.py.

All of the above methods except for iter() have asynchronous variants
as well, which return a Future; to get the operation's ultimate
result, yield the Future (when inside a tasklet) or call the Future's
get_result() method (outside a tasklet)::

  q.map_async(callback)  # Callback may be a task or a plain function
  q.fetch_async(N)
  q.get_async()
  q.count_async(N)
  q.fetch_page_async(N, start_cursor=cursor)

Finally, there's an idiom to efficiently loop over the Query results
in a tasklet, properly yielding when appropriate::

  it = q.iter()
  while (yield it.has_next_async()):
    emp = it.next()
    print emp.name, emp.age

"""



import datetime
import functools
import heapq
import sys


from google.appengine.ext.ndb import context
from google.appengine.ext.ndb import model
from google.appengine.ext.ndb import tasklets
from google.appengine.ext.ndb import utils
import six
from six.moves import map
from six.moves import range
from six.moves import zip

from google.appengine.api import cmp_compat
from google.appengine.api import datastore_errors
from google.appengine.api import datastore_types
from google.appengine.api import namespace_manager
from google.appengine.datastore import datastore_query
from google.appengine.datastore import datastore_rpc

__all__ = ['Query', 'QueryOptions', 'Cursor', 'QueryIterator',
           'RepeatedStructuredPropertyPredicate',
           'AND', 'OR', 'ConjunctionNode', 'DisjunctionNode',
           'FilterNode', 'PostFilterNode', 'FalseNode', 'Node',
           'ParameterNode', 'ParameterizedThing', 'Parameter',
           'ParameterizedFunction', 'gql',
          ]


Cursor = datastore_query.Cursor


_ASC = datastore_query.PropertyOrder.ASCENDING
_DESC = datastore_query.PropertyOrder.DESCENDING
_AND = datastore_query.CompositeFilter.AND
_KEY = datastore_types._KEY_SPECIAL_PROPERTY


_OPS = frozenset(['=', '!=', '<', '<=', '>', '>=', 'in'])


_MAX_LIMIT = 2 ** 31 - 1


class QueryOptions(context.ContextOptions, datastore_query.QueryOptions):
  """Support both context options and query options (esp. use_cache)."""


class RepeatedStructuredPropertyPredicate(datastore_query.FilterPredicate):


  def __init__(self, match_keys, pb, key_prefix):
    super(RepeatedStructuredPropertyPredicate, self).__init__()
    self.match_keys = match_keys
    stripped_keys = []
    for key in match_keys:
      if not key.startswith(key_prefix):
        raise ValueError('key %r does not begin with the specified prefix of %s'
                         % (key, key_prefix))
      stripped_key = six.ensure_text(key[len(key_prefix):])
      stripped_keys.append(stripped_key)
    value_map = datastore_query._make_key_value_map(pb, stripped_keys)
    self.match_values = tuple(value_map[key][0] for key in stripped_keys)

  def _get_prop_names(self):
    return frozenset(self.match_keys)

  def _apply(self, key_value_map):
    """Apply the filter to values extracted from an entity.

    Think of self.match_keys and self.match_values as representing a
    table with one row.  For example:

      match_keys = ('name', 'age', 'rank')
      match_values = ('Joe', 24, 5)

    (Except that in reality, the values are represented by tuples
    produced by datastore_types.PropertyValueToKeyValue().)

    represents this table:

      |  name   |  age  |  rank  |
      +---------+-------+--------+
      |  'Joe'  |   24  |     5  |

    Think of key_value_map as a table with the same structure but
    (potentially) many rows.  This represents a repeated structured
    property of a single entity.  For example:

      {'name': ['Joe', 'Jane', 'Dick'],
       'age': [24, 21, 23],
       'rank': [5, 1, 2]}

    represents this table:

      |  name   |  age  |  rank  |
      +---------+-------+--------+
      |  'Joe'  |   24  |     5  |
      |  'Jane' |   21  |     1  |
      |  'Dick' |   23  |     2  |

    We must determine wheter at least one row of the second table
    exactly matches the first table.  We need this class because the
    datastore, when asked to find an entity with name 'Joe', age 24
    and rank 5, will include entities that have 'Joe' somewhere in the
    name column, 24 somewhere in the age column, and 5 somewhere in
    the rank column, but not all aligned on a single row.  Such an
    entity should not be considered a match.
    """
    columns = []
    for key in self.match_keys:
      column = key_value_map.get(six.ensure_text(key))
      if not column:
        return False
      columns.append(column)

    return self.match_values in zip(*columns)





class ParameterizedThing(object):
  """Base class for Parameter and ParameterizedFunction.

  This exists purely for isinstance() checks.
  """

  def __eq__(self, other):
    raise NotImplementedError

  def __ne__(self, other):
    eq = self.__eq__(other)
    if eq is not NotImplemented:
      eq = not eq
    return eq


class Parameter(ParameterizedThing):
  """Represents a bound variable in a GQL query.

  Parameter(1) corresponds to a slot labeled ":1" in a GQL query.
  Parameter('xyz') corresponds to a slot labeled ":xyz".

  The value must be set (bound) separately by calling .set(value).
  """

  def __init__(self, key):
    """Constructor.

    Args:
      key: The Parameter key, must be either an integer or a string.
    """
    if not isinstance(key, six.integer_types + six.string_types):
      raise TypeError('Parameter key must be an integer or string, not %s' %
                      (key,))
    self.__key = key

  def __repr__(self):
    return '%s(%r)' % (self.__class__.__name__, self.__key)

  def __eq__(self, other):
    if not isinstance(other, Parameter):
      return NotImplemented
    return self.__key == other.__key

  @property
  def key(self):
    """Retrieve the key."""
    return self.__key

  def resolve(self, bindings, used):
    key = self.__key
    if key not in bindings:
      raise datastore_errors.BadArgumentError(
          'Parameter :%s is not bound.' % key)
    value = bindings[key]
    used[key] = True
    return value


class ParameterizedFunction(ParameterizedThing):
  """Represents a GQL function with parameterized arguments.

  For example, ParameterizedFunction('key', [Parameter(1)]) stands for
  the GQL syntax KEY(:1).
  """

  def __init__(self, func, values):


    from google.appengine.ext import gql

    self.__func = func
    self.__values = values


    gqli = gql.GQL('SELECT * FROM Dummy')
    gql_method = gqli._GQL__cast_operators[func]
    self.__method = getattr(gqli, '_GQL' + gql_method.__name__)

  def __repr__(self):
    return 'ParameterizedFunction(%r, %r)' % (self.__func, self.__values)

  def __eq__(self, other):
    if not isinstance(other, ParameterizedFunction):
      return NotImplemented
    return (self.__func == other.__func and
            self.__values == other.__values)

  @property
  def func(self):
    return self.__func

  @property
  def values(self):
    return self.__values

  def is_parameterized(self):
    for val in self.__values:
      if isinstance(val, Parameter):
        return True
    return False

  def resolve(self, bindings, used):
    values = []
    for val in self.__values:
      if isinstance(val, Parameter):
        val = val.resolve(bindings, used)
      values.append(val)
    result = self.__method(values)

    if self.__func == 'key' and isinstance(result, datastore_types.Key):
      result = model.Key.from_old_key(result)
    elif self.__func == 'time' and isinstance(result, datetime.datetime):
      result = datetime.time(result.hour, result.minute,
                             result.second, result.microsecond)
    elif self.__func == 'date' and isinstance(result, datetime.datetime):
      result = datetime.date(result.year, result.month, result.day)
    return result


class Node(object):
  """Base class for filter expression tree nodes.

  Tree nodes are considered immutable, even though they can contain
  Parameter instances, which are not.  In particular, two identical
  trees may be represented by the same Node object in different
  contexts.
  """

  def __new__(cls):
    if cls is Node:
      raise TypeError('Cannot instantiate Node, only a subclass.')
    return super(Node, cls).__new__(cls)

  def __eq__(self, other):
    raise NotImplementedError

  def __ne__(self, other):
    eq = self.__eq__(other)
    if eq is not NotImplemented:
      eq = not eq
    return eq

  def __unordered(self, unused_other):
    raise TypeError('Nodes cannot be ordered')
  __le__ = __lt__ = __ge__ = __gt__ = __unordered

  def _to_filter(self, post=False):
    """Helper to convert to datastore_query.Filter, or None."""
    raise NotImplementedError

  def _post_filters(self):
    """Helper to extract post-filter Nodes, if any."""
    return None

  def resolve(self, bindings, used):
    """Return a Node with Parameters replaced by the selected values.

    Args:
      bindings: A dict mapping integers and strings to values.
      used: A dict into which use of use of a binding is recorded.

    Returns:
      A Node instance.
    """
    return self


class FalseNode(Node):
  """Tree node for an always-failing filter."""

  def __eq__(self, other):
    if not isinstance(other, FalseNode):
      return NotImplemented
    return True

  def _to_filter(self, post=False):
    if post:
      return None


    raise datastore_errors.BadQueryError(
        'Cannot convert FalseNode to predicate')


class ParameterNode(Node):
  """Tree node for a parameterized filter."""

  def __new__(cls, prop, op, param):
    if not isinstance(prop, model.Property):
      raise TypeError('Expected a Property, got %r' % (prop,))
    if op not in _OPS:
      raise TypeError('Expected a valid operator, got %r' % (op,))
    if not isinstance(param, ParameterizedThing):
      raise TypeError('Expected a ParameterizedThing, got %r' % (param,))
    obj = super(ParameterNode, cls).__new__(cls)
    obj.__prop = prop
    obj.__op = op
    obj.__param = param
    return obj

  def __getnewargs__(self):
    return self.__prop, self.__op, self.__param

  def __repr__(self):
    return 'ParameterNode(%r, %r, %r)' % (self.__prop, self.__op, self.__param)

  def __eq__(self, other):
    if not isinstance(other, ParameterNode):
      return NotImplemented
    return (self.__prop._name == other.__prop._name and
            self.__op == other.__op and
            self.__param == other.__param)

  def _to_filter(self, post=False):
    raise datastore_errors.BadArgumentError(
        'Parameter :%s is not bound.' % (self.__param.key,))

  def resolve(self, bindings, used):
    value = self.__param.resolve(bindings, used)
    if self.__op == 'in':
      return self.__prop._IN(value)
    else:
      return self.__prop._comparison(self.__op, value)


class FilterNode(Node):
  """Tree node for a single filter expression."""

  def __new__(cls, name, opsymbol, value):
    if isinstance(value, model.Key):
      value = value.to_old_key()
    if opsymbol == '!=':
      n1 = FilterNode(name, '<', value)
      n2 = FilterNode(name, '>', value)
      return DisjunctionNode(n1, n2)
    if opsymbol == 'in':
      if not isinstance(value, (list, tuple, set, frozenset)):
        raise TypeError('in expected a list, tuple or set of values; '
                        'received %r' % value)
      nodes = [FilterNode(name, '=', v) for v in value]
      if not nodes:
        return FalseNode()
      if len(nodes) == 1:
        return nodes[0]
      return DisjunctionNode(*nodes)
    self = super(FilterNode, cls).__new__(cls)
    self.__name = name
    self.__opsymbol = opsymbol
    self.__value = value
    return self

  def __getnewargs__(self):
    return self.__name, self.__opsymbol, self.__value

  def __repr__(self):
    return '%s(%r, %r, %r)' % (self.__class__.__name__,
                               self.__name, self.__opsymbol, self.__value)

  def __eq__(self, other):
    if not isinstance(other, FilterNode):
      return NotImplemented


    return (self.__name == other.__name and
            self.__opsymbol == other.__opsymbol and
            self.__value == other.__value)

  def __hash__(self):
    return hash((self.__name, self.__opsymbol, self.__value))

  def _to_filter(self, post=False):
    if post:
      return None
    if self.__opsymbol in ('!=', 'in'):
      raise NotImplementedError('Inequality filters are not single filter '
                                'expressions and therefore cannot be converted '
                                'to a single filter (%r)' % self.__opsymbol)
    value = self.__value
    return datastore_query.make_filter(six.ensure_text(self.__name),
                                       self.__opsymbol, value)


class PostFilterNode(Node):
  """Tree node representing an in-memory filtering operation.

  This is used to represent filters that cannot be executed by the
  datastore, for example a query for a structured value.
  """

  def __new__(cls, predicate):
    self = super(PostFilterNode, cls).__new__(cls)
    self.predicate = predicate
    return self

  def __getnewargs__(self):
    return self.predicate,

  def __repr__(self):
    return '%s(%s)' % (self.__class__.__name__, self.predicate)

  def __eq__(self, other):
    if not isinstance(other, PostFilterNode):
      return NotImplemented
    return self is other or self.__dict__ == other.__dict__

  def _to_filter(self, post=False):
    if post:
      return self.predicate
    else:
      return None


class ConjunctionNode(Node):
  """Tree node representing a Boolean `AND` operator on two or more nodes."""

  def __new__(cls, *nodes):
    if not nodes:
      raise TypeError('ConjunctionNode() requires at least one node.')
    elif len(nodes) == 1:
      return nodes[0]
    clauses = [[]]

    for node in nodes:
      if not isinstance(node, Node):
        raise TypeError('ConjunctionNode() expects Node instances as arguments;'
                        ' received a non-Node instance %r' % node)
      if isinstance(node, DisjunctionNode):


        new_clauses = []
        for clause in clauses:
          for subnode in node:
            new_clause = clause + [subnode]
            new_clauses.append(new_clause)
        clauses = new_clauses
      elif isinstance(node, ConjunctionNode):


        for clause in clauses:
          clause.extend(node.__nodes)
      else:

        for clause in clauses:
          clause.append(node)
    if not clauses:
      return FalseNode()
    if len(clauses) > 1:
      return DisjunctionNode(*[ConjunctionNode(*clause) for clause in clauses])
    self = super(ConjunctionNode, cls).__new__(cls)
    self.__nodes = clauses[0]
    return self

  def __getnewargs__(self):
    return tuple(self.__nodes)

  def __iter__(self):
    return iter(self.__nodes)

  def __repr__(self):
    return 'AND(%s)' % (', '.join(map(str, self.__nodes)))

  def __eq__(self, other):
    if not isinstance(other, ConjunctionNode):
      return NotImplemented
    return self.__nodes == other.__nodes

  def _to_filter(self, post=False):

    filters = [
        node._to_filter(post=post)
        for node in self.__nodes
        if isinstance(node, PostFilterNode) == post
    ]

    filters = [f for f in filters if f]
    if not filters:
      return None
    if len(filters) == 1:
      return filters[0]
    return datastore_query.CompositeFilter(_AND, filters)

  def _post_filters(self):
    post_filters = [node for node in self.__nodes
                    if isinstance(node, PostFilterNode)]
    if not post_filters:
      return None
    if len(post_filters) == 1:
      return post_filters[0]
    if post_filters == self.__nodes:
      return self
    return ConjunctionNode(*post_filters)

  def resolve(self, bindings, used):
    nodes = [node.resolve(bindings, used) for node in self.__nodes]
    if nodes == self.__nodes:
      return self
    return ConjunctionNode(*nodes)


class DisjunctionNode(Node):
  """Tree node representing a Boolean OR operator on two or more nodes."""

  def __new__(cls, *nodes):
    if not nodes:
      raise TypeError('DisjunctionNode() requires at least one node')
    elif len(nodes) == 1:
      return nodes[0]
    self = super(DisjunctionNode, cls).__new__(cls)
    self.__nodes = []

    for node in nodes:
      if not isinstance(node, Node):
        raise TypeError('DisjunctionNode() expects Node instances as arguments;'
                        ' received a non-Node instance %r' % node)
      if isinstance(node, DisjunctionNode):
        self.__nodes.extend(node.__nodes)
      else:
        self.__nodes.append(node)
    return self

  def __getnewargs__(self):
    return tuple(self.__nodes)

  def __iter__(self):
    return iter(self.__nodes)

  def __repr__(self):
    return 'OR(%s)' % (', '.join(map(str, self.__nodes)))

  def __eq__(self, other):
    if not isinstance(other, DisjunctionNode):
      return NotImplemented
    return self.__nodes == other.__nodes

  def resolve(self, bindings, used):
    nodes = [node.resolve(bindings, used) for node in self.__nodes]
    if nodes == self.__nodes:
      return self
    return DisjunctionNode(*nodes)



AND = ConjunctionNode
OR = DisjunctionNode


def _args_to_val(func, args):
  """Helper for GQL parsing to extract values from GQL expressions.

  This can extract the value from a GQL literal, return a Parameter
  for a GQL bound parameter (:1 or :foo), and interprets casts like
  KEY(...) and plain lists of values like (1, 2, 3).

  Args:
    func: A string indicating what kind of thing this is.
    args: One or more GQL values, each integer, string, or GQL literal.
  """


  from google.appengine.ext import gql

  vals = []
  for arg in args:
    if isinstance(arg, six.integer_types + six.string_types):
      val = Parameter(arg)
    elif isinstance(arg, gql.Literal):
      val = arg.Get()
    else:
      raise TypeError('Unexpected arg (%r)' % arg)
    vals.append(val)
  if func == 'nop':
    if len(vals) != 1:
      raise TypeError('"nop" requires exactly one value')
    return vals[0]
  pfunc = ParameterizedFunction(func, vals)
  if pfunc.is_parameterized():
    return pfunc
  else:
    return pfunc.resolve({}, {})


def _get_prop_from_modelclass(modelclass, name):
  """Helper for FQL parsing to turn a property name into a property object.

  Args:
    modelclass: The model class specified in the query.
    name: The property name.  This may contain dots which indicate
      sub-properties of structured properties.

  Returns:
    A Property object.

  Raises:
    KeyError if the property doesn't exist and the model clas doesn't
    derive from Expando.
  """
  if name == '__key__':
    return modelclass._key

  parts = name.split('.')
  part, more = parts[0], parts[1:]
  prop = modelclass._properties.get(part)
  if prop is None:
    if issubclass(modelclass, model.Expando):
      prop = model.GenericProperty(part)
    else:
      raise TypeError('Model %s has no property named %r' %
                      (modelclass._get_kind(), part))

  while more:
    part = more.pop(0)
    if not isinstance(prop, model.StructuredProperty):
      raise TypeError('Model %s has no property named %r' %
                      (modelclass._get_kind(), part))
    maybe = getattr(prop, part, None)
    if isinstance(maybe, model.Property) and maybe._name == part:
      prop = maybe
    else:
      maybe = prop._modelclass._properties.get(part)
      if maybe is not None:


        prop = getattr(prop, maybe._code_name)
      else:
        if issubclass(prop._modelclass, model.Expando) and not more:
          prop = model.GenericProperty()
          prop._name = name
        else:
          raise KeyError('Model %s has no property named %r' %
                         (prop._modelclass._get_kind(), part))

  return prop


class Query(object):
  """Query object.

  Usually constructed by calling Model.query().

  See module docstring for examples.

  Note that not all operations on Queries are supported by _MultiQuery
  instances; the latter are generated as necessary when any of the
  operators !=, IN or OR is used.
  """

  @utils.positional(1)
  def __init__(self, kind=None, ancestor=None, filters=None, orders=None,
               app=None, namespace=None, default_options=None,
               projection=None, group_by=None):
    """Constructor.
    Args:
      kind: Optional kind string.
      ancestor: Optional ancestor Key.
      filters: Optional Node representing a filter expression tree.
      orders: Optional datastore_query.Order object.
      app: Optional app id.
      namespace: Optional namespace.
      default_options: Optional QueryOptions object.
      projection: Optional list or tuple of properties to project.
      group_by: Optional list or tuple of properties to group by.
    """





    if ancestor is not None:
      if isinstance(ancestor, ParameterizedThing):
        if isinstance(ancestor, ParameterizedFunction):
          if ancestor.func != 'key':
            raise TypeError('ancestor cannot be a GQL function other than KEY')
      else:
        if not isinstance(ancestor, model.Key):
          raise TypeError('ancestor must be a Key; received %r' % (ancestor,))
        if not ancestor.id():
          raise ValueError('ancestor cannot be an incomplete key')
        if app is not None:
          if app != ancestor.app():
            raise TypeError(
                'app/ancestor mismatch: %r != %r' % (app, ancestor.app()))
        if namespace is None:
          namespace = ancestor.namespace()
        else:
          if namespace != ancestor.namespace():
            raise TypeError(
                'namespace/ancestor mismatch: %r != %r' % (
                    namespace, ancestor.namespace()))
    if filters is not None:
      if not isinstance(filters, Node):
        raise TypeError('filters must be a query Node or None; received %r' %
                        (filters,))
    if orders is not None:
      if not isinstance(orders, datastore_query.Order):
        raise TypeError('orders must be an Order instance or None; received %r'
                        % (orders,))
    if default_options is not None:
      if not isinstance(default_options, datastore_rpc.BaseConfiguration):
        raise TypeError('default_options must be a Configuration or None; '
                        'received %r' % (default_options,))
      if projection is not None:
        if default_options.projection is not None:
          raise TypeError('cannot use projection= and '
                          'default_options.projection at the same time')
        if default_options.keys_only is not None:
          raise TypeError('cannot use projection= and '
                          'default_options.keys_only at the same time')

    self.__kind = kind
    self.__ancestor = ancestor
    self.__filters = filters
    self.__orders = orders
    self.__app = app
    self.__namespace = namespace
    self.__default_options = default_options


    self.__projection = None
    if projection is not None:
      if not projection:
        raise TypeError('projection argument cannot be empty')
      if not isinstance(projection, (tuple, list)):
        raise TypeError(
            'projection must be a tuple, list or None; received %r' %
            (projection,))
      self._check_properties(self._to_property_names(projection))



      self.__projection = tuple(
          six.ensure_binary(p) if isinstance(p, six.text_type) else p
          for p in projection)

    self.__group_by = None
    if group_by is not None:
      if not group_by:
        raise TypeError('group_by argument cannot be empty')
      if not isinstance(group_by, (tuple, list)):
        raise TypeError(
            'group_by must be a tuple, list or None; received %r' % (group_by,))
      self._check_properties(self._to_property_names(group_by))



      self.__group_by = tuple(
          six.ensure_binary(g) if isinstance(g, six.text_type) else g
          for g in group_by)

  def __repr__(self):
    args = []
    if self.app is not None:
      args.append('app=%r' % self.app)
    if (self.namespace is not None and
        self.namespace != namespace_manager.get_namespace()):


      args.append('namespace=%r' % self.namespace)
    if self.kind is not None:
      args.append('kind=%r' % self.kind)
    if self.ancestor is not None:
      args.append('ancestor=%r' % self.ancestor)
    if self.filters is not None:
      args.append('filters=%r' % self.filters)
    if self.orders is not None:

      args.append('orders=...')
    if self.projection:
      args.append('projection=%r' % (self._to_property_names(self.projection)))
    if self.group_by:
      args.append('group_by=%r' % (self._to_property_names(self.group_by)))
    if self.default_options is not None:
      args.append('default_options=%r' % self.default_options)
    return '%s(%s)' % (self.__class__.__name__, ', '.join(args))

  def _fix_namespace(self):
    """Internal helper to fix the namespace.

    This is called to ensure that for queries without an explicit
    namespace, the namespace used by async calls is the one in effect
    at the time the async call is made, not the one in effect when the
    the request is actually generated.
    """
    if self.namespace is not None:
      return self
    namespace = namespace_manager.get_namespace()
    return self.__class__(kind=self.kind, ancestor=self.ancestor,
                          filters=self.filters, orders=self.orders,
                          app=self.app, namespace=namespace,
                          default_options=self.default_options,
                          projection=self.projection, group_by=self.group_by)

  def _get_query(self, connection):
    self.bind()
    kind = self.kind
    ancestor = self.ancestor
    if ancestor is not None:
      ancestor = connection.adapter.key_to_pb(ancestor)
    filters = self.filters
    post_filters = None
    if filters is not None:
      post_filters = filters._post_filters()
      filters = filters._to_filter()
    group_by = None
    if self.group_by:
      group_by = self._to_property_names(self.group_by)
    dsquery = datastore_query.Query(
        app=self.app,
        namespace=self.namespace,
        kind=six.ensure_text(kind) if kind else None,
        ancestor=ancestor,
        filter_predicate=filters,
        order=self.orders,
        group_by=group_by)
    if post_filters is not None:
      dsquery = datastore_query._AugmentedQuery(
          dsquery,
          in_memory_filter=post_filters._to_filter(post=True))
    return dsquery

  @tasklets.tasklet
  def run_to_queue(self, queue, conn, options=None, dsquery=None):
    """Run this query, putting entities into the given queue."""
    try:
      multiquery = self._maybe_multi_query()
      if multiquery is not None:
        yield multiquery.run_to_queue(queue, conn, options=options)
        return

      if dsquery is None:
        dsquery = self._get_query(conn)
      rpc = dsquery.run_async(conn, options)
      while rpc is not None:
        batch = yield rpc
        if (batch.skipped_results and
            datastore_query.FetchOptions.offset(options)):
          offset = options.offset - batch.skipped_results
          options = datastore_query.FetchOptions(offset=offset, config=options)
        rpc = batch.next_batch_async(options)
        for i, result in enumerate(batch.results):
          queue.putq((batch, i, result))
      queue.complete()

    except GeneratorExit:
      raise
    except Exception:
      if not queue.done():
        _, e, tb = sys.exc_info()
        queue.set_exception(e, tb)
      raise

  @tasklets.tasklet
  def _run_to_list(self, results, options=None):

    ctx = tasklets.get_context()
    conn = ctx._conn
    dsquery = self._get_query(conn)
    rpc = dsquery.run_async(conn, options)
    while rpc is not None:
      batch = yield rpc
      if (batch.skipped_results and
          datastore_query.FetchOptions.offset(options)):
        offset = options.offset - batch.skipped_results
        options = datastore_query.FetchOptions(offset=offset, config=options)
      rpc = batch.next_batch_async(options)
      for result in batch.results:
        result = ctx._update_cache_from_query_result(result, options)
        if result is not None:
          results.append(result)

    raise tasklets.Return(results)

  def _needs_multi_query(self):
    filters = self.filters
    return filters is not None and isinstance(filters, DisjunctionNode)

  def _maybe_multi_query(self):
    if not self._needs_multi_query():
      return None

    filters = self.filters
    subqueries = []
    for subfilter in filters:
      subquery = self.__class__(kind=self.kind, ancestor=self.ancestor,
                                filters=subfilter, orders=self.orders,
                                app=self.app, namespace=self.namespace,
                                default_options=self.default_options,
                                projection=self.projection,
                                group_by=self.group_by)
      subqueries.append(subquery)
    return _MultiQuery(subqueries)

  @property
  def kind(self):
    """Accessor for the kind (a string or None)."""
    return self.__kind

  @property
  def ancestor(self):
    """Accessor for the ancestor (a Key or None)."""
    return self.__ancestor

  @property
  def filters(self):
    """Accessor for the filters (a Node or None)."""
    return self.__filters

  @property
  def orders(self):
    """Accessor for the filters (a datastore_query.Order or None)."""
    return self.__orders

  @property
  def app(self):
    """Accessor for the app (a string or None)."""
    return self.__app

  @property
  def namespace(self):
    """Accessor for the namespace (a string or None)."""
    return self.__namespace

  @property
  def default_options(self):
    """Accessor for the default_options (a QueryOptions instance or None)."""
    return self.__default_options

  @property
  def group_by(self):
    """Accessor for the group by properties (a tuple instance or None)."""
    return self.__group_by

  @property
  def projection(self):
    """Accessor for the projected properties (a tuple instance or None)."""
    return self.__projection

  @property
  def is_distinct(self):
    """True if results are guaranteed to contain a unique set of property
    values.

    This happens when every property in the group_by is also in the projection.
    """
    return bool(self.__group_by and
                set(self._to_property_names(self.__group_by)) <=
                set(self._to_property_names(self.__projection)))

  def filter(self, *args):
    """Return a new Query with additional filter(s) applied."""
    if not args:
      return self
    preds = []
    f = self.filters
    if f:
      preds.append(f)
    for arg in args:
      if not isinstance(arg, Node):
        raise TypeError('Cannot filter a non-Node argument; received %r' % arg)
      preds.append(arg)
    if not preds:
      pred = None
    elif len(preds) == 1:
      pred = preds[0]
    else:
      pred = ConjunctionNode(*preds)
    return self.__class__(kind=self.kind, ancestor=self.ancestor,
                          filters=pred, orders=self.orders,
                          app=self.app, namespace=self.namespace,
                          default_options=self.default_options,
                          projection=self.projection, group_by=self.group_by)

  def order(self, *args):
    """Return a new Query with additional sort order(s) applied."""

    if not args:
      return self
    orders = []
    o = self.orders
    if o:
      orders.append(o)
    for arg in args:
      if isinstance(arg, model.Property):
        orders.append(datastore_query.PropertyOrder(arg._name, _ASC))
      elif isinstance(arg, datastore_query.Order):
        orders.append(arg)
      else:
        raise TypeError('order() expects a Property or query Order; '
                        'received %r' % arg)
    if not orders:
      orders = None
    elif len(orders) == 1:
      orders = orders[0]
    else:
      orders = datastore_query.CompositeOrder(orders)
    return self.__class__(kind=self.kind, ancestor=self.ancestor,
                          filters=self.filters, orders=orders,
                          app=self.app, namespace=self.namespace,
                          default_options=self.default_options,
                          projection=self.projection, group_by=self.group_by)



  def iter(self, **q_options):
    """Construct an iterator over the query.

    Args:
      **q_options: All query options keyword arguments are supported.

    Returns:
      A QueryIterator object.
    """
    self.bind()
    return QueryIterator(self, **q_options)

  __iter__ = iter

  @utils.positional(2)
  def map(self, callback, pass_batch_into_callback=None,
          merge_future=None, **q_options):
    """Map a callback function or tasklet over the query results.

    Args:
      callback: A function or tasklet to be applied to each result; see below.
      merge_future: Optional Future subclass; see below.
      **q_options: All query options keyword arguments are supported.

    Callback signature: The callback is normally called with an entity
    as argument.  However if keys_only=True is given, it is called
    with a Key.  Also, when pass_batch_into_callback is True, it is
    called with three arguments: the current batch, the index within
    the batch, and the entity or Key at that index.  The callback can
    return whatever it wants.  If the callback is None, a trivial
    callback is assumed that just returns the entity or key passed in
    (ignoring produce_cursors).

    Optional merge future: The merge_future is an advanced argument
    that can be used to override how the callback results are combined
    into the overall map() return value.  By default a list of
    callback return values is produced.  By substituting one of a
    small number of specialized alternatives you can arrange
    otherwise.  See tasklets.MultiFuture for the default
    implementation and a description of the protocol the merge_future
    object must implement the default.  Alternatives from the same
    module include QueueFuture, SerialQueueFuture and ReducingFuture.

    Returns:
      When the query has run to completion and all callbacks have
      returned, map() returns a list of the results of all callbacks.
      (But see 'optional merge future' above.)
    """
    return self.map_async(callback,
                          pass_batch_into_callback=pass_batch_into_callback,
                          merge_future=merge_future,
                          **q_options).get_result()

  @utils.positional(2)
  def map_async(self, callback, pass_batch_into_callback=None,
                merge_future=None, **q_options):
    """Map a callback function or tasklet over the query results.

    This is the asynchronous version of Query.map().
    """
    qry = self._fix_namespace()
    return tasklets.get_context().map_query(
        qry,
        callback,
        pass_batch_into_callback=pass_batch_into_callback,
        options=self._make_options(q_options),
        merge_future=merge_future)

  @utils.positional(2)
  def fetch(self, limit=None, **q_options):
    """Fetch a list of query results, up to a limit.

    Args:
      limit: How many results to retrieve at most.
      **q_options: All query options keyword arguments are supported.

    Returns:
      A list of results.
    """
    return self.fetch_async(limit, **q_options).get_result()

  @utils.positional(2)
  def fetch_async(self, limit=None, **q_options):
    """Fetch a list of query results, up to a limit.

    This is the asynchronous version of Query.fetch().
    """
    if limit is None:
      default_options = self._make_options(q_options)
      if default_options is not None and default_options.limit is not None:
        limit = default_options.limit
      else:
        limit = _MAX_LIMIT
    q_options['limit'] = limit
    q_options.setdefault('batch_size', limit)
    if self._needs_multi_query():
      return self.map_async(None, **q_options)

    options = self._make_options(q_options)
    qry = self._fix_namespace()
    return qry._run_to_list([], options=options)

  def get(self, **q_options):
    """Get the first query result, if any.

    This is similar to calling q.fetch(1) and returning the first item
    of the list of results, if any, otherwise None.

    Args:
      **q_options: All query options keyword arguments are supported.

    Returns:
      A single result, or None if there are no results.
    """
    return self.get_async(**q_options).get_result()

  def get_async(self, **q_options):
    """Get the first query result, if any.

    This is the asynchronous version of Query.get().
    """
    qry = self._fix_namespace()
    return qry._get_async(**q_options)

  @tasklets.tasklet
  def _get_async(self, **q_options):
    """Internal version of get_async()."""
    res = yield self.fetch_async(1, **q_options)
    if not res:
      raise tasklets.Return(None)
    raise tasklets.Return(res[0])

  @utils.positional(2)
  def count(self, limit=None, **q_options):
    """Count the number of query results, up to a limit.

    This returns the same result as len(q.fetch(limit)) but more
    efficiently.

    Note that you must pass a maximum value to limit the amount of
    work done by the query.

    Args:
      limit: How many results to count at most.
      **q_options: All query options keyword arguments are supported.

    Returns:
    """
    return self.count_async(limit, **q_options).get_result()

  @utils.positional(2)
  def count_async(self, limit=None, **q_options):
    """Count the number of query results, up to a limit.

    This is the asynchronous version of Query.count().
    """
    qry = self._fix_namespace()
    return qry._count_async(limit=limit, **q_options)

  @tasklets.tasklet
  def _count_async(self, limit=None, **q_options):
    """Internal version of count_async()."""

    if 'offset' in q_options:
      raise NotImplementedError('.count() and .count_async() do not support '
                                'offsets at present.')
    if 'limit' in q_options:
      raise TypeError('Cannot specify limit as a non-keyword argument and as a '
                      'keyword argument simultaneously.')
    elif limit is None:
      limit = _MAX_LIMIT
    if self._needs_multi_query():



      q_options.setdefault('batch_size', limit)
      q_options.setdefault('keys_only', True)
      results = yield self.fetch_async(limit, **q_options)
      raise tasklets.Return(len(results))




    q_options['offset'] = limit
    q_options['limit'] = 0
    options = self._make_options(q_options)
    conn = tasklets.get_context()._conn
    dsquery = self._get_query(conn)
    rpc = dsquery.run_async(conn, options)
    total = 0
    while rpc is not None:
      batch = yield rpc
      options = QueryOptions(offset=options.offset - batch.skipped_results,
                             config=options)
      rpc = batch.next_batch_async(options)
      total += batch.skipped_results
    raise tasklets.Return(total)

  @utils.positional(2)
  def fetch_page(self, page_size, **q_options):
    """Fetch a page of results.

    This is a specialized method for use by paging user interfaces.

    Args:
      page_size: The requested page size.  At most this many results
        will be returned.

    In addition, any keyword argument supported by the QueryOptions
    class is supported.  In particular, to fetch the next page, you
    pass the cursor returned by one call to the next call using
    start_cursor=<cursor>.  A common idiom is to pass the cursor to
    the client using <cursor>.to_websafe_string() and to reconstruct
    that cursor on a subsequent request using
    Cursor.from_websafe_string(<string>).

    Returns:
      A tuple (results, cursor, more) where results is a list of query
      results, cursor is a cursor pointing just after the last result
      returned, and more is a bool indicating whether there are
      (likely) more results after that.
    """

    return self.fetch_page_async(page_size, **q_options).get_result()

  @utils.positional(2)
  def fetch_page_async(self, page_size, **q_options):
    """Fetch a page of results.

    This is the asynchronous version of Query.fetch_page().
    """
    qry = self._fix_namespace()
    return qry._fetch_page_async(page_size, **q_options)

  @tasklets.tasklet
  def _fetch_page_async(self, page_size, **q_options):
    """Internal version of fetch_page_async()."""
    q_options.setdefault('batch_size', page_size)
    q_options.setdefault('produce_cursors', True)
    it = self.iter(limit=page_size + 1, **q_options)
    results = []
    while (yield it.has_next_async()):
      results.append(next(it))
      if len(results) >= page_size:
        break
    try:
      cursor = it.cursor_after()
    except datastore_errors.BadArgumentError:
      cursor = None
    raise tasklets.Return(results, cursor, it.probably_has_next())

  def _make_options(self, q_options):
    """Helper to construct a QueryOptions object from keyword arguments.

    Args:
      q_options: a dict of keyword arguments.

    Note that either 'options' or 'config' can be used to pass another
    QueryOptions object, but not both.  If another QueryOptions object is
    given it provides default values.

    If self.default_options is set, it is used to provide defaults,
    which have a lower precedence than options set in q_options.

    Returns:
      A QueryOptions object, or None if q_options is empty.
    """
    if not (q_options or self.__projection):
      return self.default_options
    if 'options' in q_options:

      if 'config' in q_options:
        raise TypeError('You cannot use config= and options= at the same time')
      q_options['config'] = q_options.pop('options')
    if q_options.get('projection'):
      try:
        q_options['projection'] = self._to_property_names(
            q_options['projection'])
      except TypeError as e:
        raise datastore_errors.BadArgumentError(e)
      self._check_properties(q_options['projection'])
    options = QueryOptions(**q_options)


    if (options.keys_only is None and
        options.projection is None and
        self.__projection):
      options = QueryOptions(
          projection=self._to_property_names(self.__projection), config=options)

    if self.default_options is not None:
      options = self.default_options.merge(options)

    return options

  def _to_property_names(self, properties):
    if not isinstance(properties, (list, tuple)):
      properties = [properties]
    fixed = []
    for proj in properties:
      if isinstance(proj, (six.text_type, six.binary_type)):
        fixed.append(proj)
      elif isinstance(proj, model.Property):
        fixed.append(proj._name)
      else:
        raise TypeError(
            'Unexpected property (%r); should be string or Property' % (proj,))
    return fixed

  def _check_properties(self, fixed, **kwargs):
    modelclass = model.Model._kind_map.get(self.__kind)
    if modelclass is not None:
      modelclass._check_properties(fixed, **kwargs)

  def analyze(self):
    """Return a list giving the parameters required by a query."""
    class MockBindings(dict):

      def __contains__(self, key):
        self[key] = None
        return True
    bindings = MockBindings()
    used = {}
    ancestor = self.ancestor
    if isinstance(ancestor, ParameterizedThing):
      ancestor = ancestor.resolve(bindings, used)
    filters = self.filters
    if filters is not None:
      filters = filters.resolve(bindings, used)



    if six.PY2:
      return sorted(used)
    else:
      return sorted(used, key=functools.cmp_to_key(cmp_compat.cmp))

  def bind(self, *args, **kwds):
    """Bind parameter values.  Returns a new Query object."""
    return self._bind(args, kwds)

  def _bind(self, args, kwds):
    """Bind parameter values.  Returns a new Query object."""
    bindings = dict(kwds)
    for i, arg in enumerate(args):
      bindings[i + 1] = arg
    used = {}
    ancestor = self.ancestor
    if isinstance(ancestor, ParameterizedThing):
      ancestor = ancestor.resolve(bindings, used)
    filters = self.filters
    if filters is not None:
      filters = filters.resolve(bindings, used)
    unused = []
    for i in range(1, 1 + len(args)):
      if i not in used:
        unused.append(i)
    if unused:
      raise datastore_errors.BadArgumentError(
          'Positional arguments %s were given but not used.' %
          ', '.join(str(i) for i in unused))
    return self.__class__(kind=self.kind, ancestor=ancestor,
                          filters=filters, orders=self.orders,
                          app=self.app, namespace=self.namespace,
                          default_options=self.default_options,
                          projection=self.projection, group_by=self.group_by)


def gql(query_string, *args, **kwds):
  """Parse a GQL query string.

  Args:
    query_string: Full GQL query, e.g. 'SELECT * FROM Kind WHERE prop = 1'.
    *args, **kwds: If present, used to call bind().

  Returns:
    An instance of query_class.
  """
  qry = _gql(query_string)
  if args or kwds:
    qry = qry._bind(args, kwds)
  return qry


@utils.positional(1)
def _gql(query_string, query_class=Query):
  """Parse a GQL query string (internal version).

  Args:
    query_string: Full GQL query, e.g. 'SELECT * FROM Kind WHERE prop = 1'.
    query_class: Optional class to use, default Query.

  Returns:
    An instance of query_class.
  """


  from google.appengine.ext import gql

  gql_qry = gql.GQL(query_string)
  kind = gql_qry.kind()
  if kind is None:



    modelclass = model.Expando
  else:
    modelclass = model.Model._lookup_model(
        kind,
        tasklets.get_context()._conn.adapter.default_model)

    kind = modelclass._get_kind()
  ancestor = None
  flt = gql_qry.filters()
  filters = list(modelclass._default_filters())
  for name_op in sorted(flt, key=functools.cmp_to_key(cmp_compat.cmp)):
    name, op = name_op
    values = flt[name_op]
    op = op.lower()
    if op == 'is' and name == gql.GQL._GQL__ANCESTOR:
      if len(values) != 1:
        raise ValueError('"is" requires exactly one value')
      [(func, args)] = values
      ancestor = _args_to_val(func, args)
      continue
    if op not in _OPS:
      raise NotImplementedError('Operation %r is not supported.' % op)
    for (func, args) in values:
      val = _args_to_val(func, args)
      prop = _get_prop_from_modelclass(modelclass, name)
      if six.ensure_text(prop._name) != name:
        raise RuntimeError('Whoa! _get_prop_from_modelclass(%s, %r) '
                           'returned a property whose name is %r?!' %
                           (modelclass.__name__, name, prop._name))
      if isinstance(val, ParameterizedThing):
        node = ParameterNode(prop, op, val)
      elif op == 'in':
        node = prop._IN(val)
      else:
        node = prop._comparison(op, val)
      filters.append(node)
  if filters:
    filters = ConjunctionNode(*filters)
  else:
    filters = None
  orders = _orderings_to_orders(gql_qry.orderings(), modelclass)
  offset = gql_qry.offset()
  limit = gql_qry.limit()
  if limit < 0:
    limit = None
  keys_only = gql_qry._keys_only
  if not keys_only:
    keys_only = None
  options = QueryOptions(offset=offset, limit=limit, keys_only=keys_only)
  projection = gql_qry.projection()
  if gql_qry.is_distinct():
    group_by = projection
  else:
    group_by = None
  qry = query_class(kind=kind,
                    ancestor=ancestor,
                    filters=filters,
                    orders=orders,
                    default_options=options,
                    projection=projection,
                    group_by=group_by)
  return qry


class QueryIterator(six.Iterator):
  """This iterator works both for synchronous and async callers!

  For synchronous callers, just use:

    for entity in Account.query():
      <use entity>

  Async callers use this idiom:

    it = iter(Account.query())
    while (yield it.has_next_async()):
      entity = it.next()
      <use entity>

  You can also use q.iter([options]) instead of iter(q); this allows
  passing query options such as keys_only or produce_cursors.

  When keys_only is set, it.next() returns a key instead of an entity.

  When produce_cursors is set, the methods it.cursor_before() and
  it.cursor_after() return Cursor objects corresponding to the query
  position just before and after the item returned by it.next().
  Before it.next() is called for the first time, both raise an
  exception.  Once the loop is exhausted, both return the cursor after
  the last item returned.  Calling it.has_next() does not affect the
  cursors; you must call it.next() before the cursors move.  Note that
  sometimes requesting a cursor requires a Cloud Datastore roundtrip (but
  not if you happen to request a cursor corresponding to a batch
  boundary).  If produce_cursors is not set, both methods always raise
  an exception.

  Note that queries requiring in-memory merging of multiple queries
  (i.e. queries using the IN, != or OR operators) do not support query
  options.
  """


  _cursor_before = _cursor_after = (
      datastore_errors.BadArgumentError('There is no cursor currently'))


  _index_list = None



  _more_results = None


  _exhausted = False

  @utils.positional(2)
  def __init__(self, query, **q_options):
    """Constructor.  Takes a Query and query options.

    This is normally called by Query.iter() or Query.__iter__().
    """
    ctx = tasklets.get_context()
    options = query._make_options(q_options)
    callback = self._extended_callback
    self._iter = ctx.iter_query(query,
                                callback=callback,
                                pass_batch_into_callback=True,
                                options=options)
    self._fut = None

  def _extended_callback(self, batch, index, ent):
    if self._exhausted:
      raise RuntimeError('QueryIterator is already exhausted')



    if batch is None:
      return (ent, None, None, None)

    if self._index_list is None:
      self._index_list = getattr(batch, 'index_list', None)



    more_results = index + 1 < len(batch.results) or batch.more_results

    before_cursor = None
    after_cursor = None
    try:
      before_cursor = batch.cursor(index)
    except BaseException as e:
      before_cursor = e
    try:
      after_cursor = batch.cursor(index + 1)
    except BaseException as e:
      after_cursor = e
    return (ent, before_cursor, after_cursor, more_results)

  def cursor_before(self):
    """Return the cursor before the current item.

    You must pass a QueryOptions object with produce_cursors=True
    for this to work.

    If there is no cursor or no current item, raise BadArgumentError.
    Before next() has returned there is no cursor.  Once the loop is
    exhausted, this returns the cursor after the last item.
    """
    if self._exhausted:
      return self.cursor_after()
    if isinstance(self._cursor_before, BaseException):
      raise self._cursor_before
    return self._cursor_before

  def cursor_after(self):
    """Return the cursor after the current item.

    You must pass a QueryOptions object with produce_cursors=True
    for this to work.

    If there is no cursor or no current item, raise BadArgumentError.
    Before next() has returned there is no cursor.    Once the loop is
    exhausted, this returns the cursor after the last item.
    """
    if isinstance(self._cursor_after, BaseException):
      raise self._cursor_after
    return self._cursor_after

  def index_list(self):
    """Return the list of indexes used for this query.

    This returns a list of index representations, where an index
    representation is the same as what is returned by get_indexes().

    Before the first result, the information is unavailable, and then
    None is returned.  This is not the same as an empty list -- the
    empty list means that no index was used to execute the query.  (In
    the dev_appserver, an empty list may also mean that only built-in
    indexes were used; metadata queries also return an empty list
    here.)

    Proper use is as follows:
      q = <modelclass>.query(<filters>)
      i = q.iter()
      try:
        i.next()
      except Stopiteration:
        pass
      indexes = i.index_list()
      assert isinstance(indexes, list)

    Notes:
    - Forcing produce_cursors=False makes this always return None.
    - This always returns None for a multi-query.
    """




    return self._index_list


  def __iter__(self):
    """Iterator protocol: get the iterator for this iterator, i.e. self."""
    return self


  def probably_has_next(self):
    """Return whether a next item is (probably) available.

     This is not quite the same as has_next(), because when
    produce_cursors is set, some shortcuts are possible.  However, in
    some cases (e.g. when the query has a post_filter) we can get a
    false positive (returns True but next() will raise StopIteration).
    There are no false negatives."""




    if self._more_results:
      return True
    return self.has_next()

  def has_next(self):
    """Return whether a next item is available.

    See the module docstring for the usage pattern.
    """
    return self.has_next_async().get_result()

  @tasklets.tasklet
  def has_next_async(self):
    """Return a Future whose result will say whether a next item is available.

    See the module docstring for the usage pattern.
    """
    if self._fut is None:
      self._fut = self._iter.getq()
    flag = True
    try:
      yield self._fut
    except EOFError:
      flag = False
    raise tasklets.Return(flag)

  def __next__(self):
    """Iterator protocol: get next item or raise StopIteration."""
    if self._fut is None:
      self._fut = self._iter.getq()
    try:
      try:



        (ent,
         self._cursor_before,
         self._cursor_after,
         self._more_results) = self._fut.get_result()
        return ent
      except EOFError:
        self._exhausted = True
        raise StopIteration
    finally:
      self._fut = None


@cmp_compat.total_ordering_from_cmp
class _SubQueryIteratorState(object):
  """Helper class for _MultiQuery."""

  def __init__(self, batch_i_entity, iterator, dsquery, orders):
    batch, index, entity = batch_i_entity
    self.batch = batch
    self.index = index
    self.entity = entity
    self.iterator = iterator
    self.dsquery = dsquery
    self.orders = orders

  def __cmp__(self, other):
    if not isinstance(other, _SubQueryIteratorState):
      raise NotImplementedError('Can only compare _SubQueryIteratorState '
                                'instances to other _SubQueryIteratorState '
                                'instances; not %r' % other)
    if not self.orders == other.orders:
      raise NotImplementedError('Cannot compare _SubQueryIteratorStates with '
                                'differing orders (%r != %r)' %
                                (self.orders, other.orders))
    lhs = self.entity._orig_pb
    rhs = other.entity._orig_pb
    lhs_filter = self.dsquery.filter_predicate
    rhs_filter = other.dsquery.filter_predicate
    names = self.orders._get_prop_names()


    if lhs_filter is not None:
      names |= lhs_filter._get_prop_names()
    if rhs_filter is not None:
      names |= rhs_filter._get_prop_names()
    lhs_value_map = datastore_query._make_key_value_map(lhs, names)
    rhs_value_map = datastore_query._make_key_value_map(rhs, names)
    if lhs_filter is not None:
      lhs_filter._prune(lhs_value_map)
    if rhs_filter is not None:
      rhs_filter._prune(rhs_value_map)
    return self.orders._cmp(lhs_value_map, rhs_value_map)


class _MultiQuery(object):
  """Helper class to run queries involving !=, IN or OR operators."""












  def __init__(self, subqueries):
    if not isinstance(subqueries, list):
      raise TypeError('subqueries must be a list; received %r' % subqueries)
    for subq in subqueries:
      if not isinstance(subq, Query):
        raise TypeError('Each subquery must be a Query instances; received  %r'
                        % subq)
    first_subquery = subqueries[0]
    kind = first_subquery.kind
    orders = first_subquery.orders
    if not kind:
      raise ValueError('Subquery kind cannot be missing')
    for subq in subqueries[1:]:
      if subq.kind != kind:
        raise ValueError('Subqueries must be for a common kind (%s != %s)' %
                         (subq.kind, kind))
      elif subq.orders != orders:
        raise ValueError('Subqueries must have the same order(s) (%s != %s)' %
                         (subq.orders, orders))

    self.__subqueries = subqueries
    self.__orders = orders
    self.ancestor = None

  def _make_options(self, q_options):
    return self.__subqueries[0].default_options

  @property
  def orders(self):
    return self.__orders

  @property
  def default_options(self):
    return self.__subqueries[0].default_options

  @tasklets.tasklet
  def run_to_queue(self, queue, conn, options=None):
    """Run this query, putting entities into the given queue."""
    if options is None:

      offset = None
      limit = None
      keys_only = None
    else:

      offset = options.offset
      limit = options.limit
      keys_only = options.keys_only


      if (options.start_cursor or options.end_cursor or
          options.produce_cursors):
        names = set()
        if self.__orders is not None:
          names = self.__orders._get_prop_names()
        if '__key__' not in names:
          raise datastore_errors.BadArgumentError(
              '_MultiQuery with cursors requires __key__ order')








    modifiers = {}
    if offset:
      modifiers['offset'] = None
      if limit is not None:
        modifiers['limit'] = min(_MAX_LIMIT, offset + limit)
    if keys_only and self.__orders is not None:
      modifiers['keys_only'] = None
    if modifiers:
      options = QueryOptions(config=options, **modifiers)

    if offset is None:
      offset = 0

    if limit is None:
      limit = _MAX_LIMIT

    if self.__orders is None:

      keys_seen = set()
      for subq in self.__subqueries:
        if limit <= 0:
          break
        subit = tasklets.SerialQueueFuture('_MultiQuery.run_to_queue[ser]')
        subq.run_to_queue(subit, conn, options=options)
        while limit > 0:
          try:
            batch, index, result = yield subit.getq()
          except EOFError:
            break
          if keys_only:
            key = result
          else:
            key = result._key
          if key not in keys_seen:
            keys_seen.add(key)
            if offset > 0:
              offset -= 1
            else:
              limit -= 1
              queue.putq((None, None, result))
      queue.complete()
      return




    with conn.adapter:

      todo = []
      for subq in self.__subqueries:
        dsquery = subq._get_query(conn)
        subit = tasklets.SerialQueueFuture('_MultiQuery.run_to_queue[par]')
        subq.run_to_queue(subit, conn, options=options, dsquery=dsquery)
        todo.append((subit, dsquery))


      state = []
      for subit, dsquery in todo:
        try:
          thing = yield subit.getq()
        except EOFError:
          continue
        else:
          state.append(_SubQueryIteratorState(thing, subit, dsquery,
                                              self.__orders))




      heapq.heapify(state)










      keys_seen = set()
      while state and limit > 0:
        item = heapq.heappop(state)
        batch = item.batch
        index = item.index
        entity = item.entity
        key = entity._key
        if key not in keys_seen:
          keys_seen.add(key)
          if offset > 0:
            offset -= 1
          else:
            limit -= 1
            if keys_only:
              queue.putq((batch, index, key))
            else:
              queue.putq((batch, index, entity))
        subit = item.iterator
        try:
          batch, index, entity = yield subit.getq()
        except EOFError:
          pass
        else:
          item.batch = batch
          item.index = index
          item.entity = entity
          heapq.heappush(state, item)
      queue.complete()



  def iter(self, **q_options):
    return QueryIterator(self, **q_options)

  __iter__ = iter








def _order_to_ordering(order):
  pb = order._to_pb()
  return pb.property, pb.direction


def _orders_to_orderings(orders):
  if orders is None:
    return []
  if isinstance(orders, datastore_query.PropertyOrder):
    return [_order_to_ordering(orders)]
  if isinstance(orders, datastore_query.CompositeOrder):

    return [(pb.property, pb.direction) for pb in orders._to_pbs()]
  raise ValueError('Bad order: %r' % (orders,))


def _ordering_to_order(ordering, modelclass):
  name, direction = ordering
  prop = _get_prop_from_modelclass(modelclass, name)
  if six.ensure_text(prop._name) != name:
    raise RuntimeError('Whoa! _get_prop_from_modelclass(%s, %r) '
                       'returned a property whose name is %r?!' %
                       (modelclass.__name__, name, prop._name))
  return datastore_query.PropertyOrder(name, direction)


def _orderings_to_orders(orderings, modelclass):
  orders = [_ordering_to_order(o, modelclass) for o in orderings]
  if not orders:
    return None
  if len(orders) == 1:
    return orders[0]
  return datastore_query.CompositeOrder(orders)



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""PyYAML event listener.

Contains class which interprets YAML events and forwards them to
a handler object.
"""


import copy
from google.appengine.api import yaml_errors
from ruamel import yaml



_EVENT_METHOD_MAP = {
  yaml.events.StreamStartEvent: 'StreamStart',
  yaml.events.StreamEndEvent: 'StreamEnd',
  yaml.events.DocumentStartEvent: 'DocumentStart',
  yaml.events.DocumentEndEvent: 'DocumentEnd',
  yaml.events.AliasEvent: 'Alias',
  yaml.events.ScalarEvent: 'Scalar',
  yaml.events.SequenceStartEvent: 'SequenceStart',
  yaml.events.SequenceEndEvent: 'SequenceEnd',
  yaml.events.MappingStartEvent: 'MappingStart',
  yaml.events.MappingEndEvent: 'MappingEnd',
}


class EventHandler(object):
  """Handler interface for parsing YAML files.

  Implement this interface to define specific YAML event handling class.
  Implementing classes instances are passed to the constructor of
  `EventListener` to act as a receiver of YAML parse events.
  """
  def StreamStart(self, event, loader):
    """Handles start of stream event."""

  def StreamEnd(self, event, loader):
    """Handles end of stream event."""

  def DocumentStart(self, event, loader):
    """Handles start of document event."""

  def DocumentEnd(self, event, loader):
    """Handles end of document event."""

  def Alias(self, event, loader):
    """Handles alias event."""

  def Scalar(self, event, loader):
    """Handles scalar event."""

  def SequenceStart(self, event, loader):
    """Handles start of sequence event."""

  def SequenceEnd(self, event, loader):
    """Handles end of sequence event."""

  def MappingStart(self, event, loader):
    """Handles start of mapping event."""

  def MappingEnd(self, event, loader):
    """Handles end of mapping event."""


class EventListener(object):
  """Helper class to re-map PyYAML events to method calls.

  By default, PyYAML generates its events via a Python generator. This class
  is a helper that iterates over the events from the PyYAML parser and forwards
  them to a handle class in the form of method calls. For simplicity, the
  underlying event is forwarded to the handler as a parameter to the call.

  This object does not produce iterable objects itself, but is instead a
  mapping to a given handler instance.

    Example use:

      ```python
      class PrintDocumentHandler(object):
        def DocumentStart(event):
          print "A new document has been started"

      EventListener(PrintDocumentHandler()).Parse('''
        key1: value1
        ---
        key2: value2
        '''

      >>> A new document has been started
          A new document has been started
      ```

  In the example above, the implemented handler class (`PrintDocumentHandler`)
  has a single method that reports each time a new document is started within a
  YAML file.  It is not necessary to subclass the `EventListener`, since only it
  receives a `PrintDocumentHandler` instance. Every time a new document begins,
  `PrintDocumentHandler.DocumentStart` is called with the PyYAML event passed in
  as its parameter.

  The constructor initializes the PyYAML event listener and constructs internal
  mapping directly from event type to method on the actual handler. This
  prevents reflection being used during the actual parse time.
  """

  def __init__(self, event_handler):
    """Constructor.

    Args:
      event_handler: Event handler that will receive mapped events. Must
        implement at least one appropriate handler method named from
        the values of the `_EVENT_METHOD_MAP`.

    Raises:
      `ListenerConfigurationError` if `event_handler` is not an `EventHandler`.
    """
    if not isinstance(event_handler, EventHandler):
      raise yaml_errors.ListenerConfigurationError(
        'Must provide event handler of type yaml_listener.EventHandler')
    self._event_method_map = {}

    for event, method in _EVENT_METHOD_MAP.items():

      self._event_method_map[event] = getattr(event_handler, method)

  def HandleEvent(self, event, loader=None):
    """Handles individual PyYAML event.

    Args:
      event: Event to forward to method call in method call.

    Raises:
      `IllegalEvent` when receives an unrecognized or unsupported event type.
    """

    if event.__class__ not in _EVENT_METHOD_MAP:
      raise yaml_errors.IllegalEvent(
            "%s is not a valid PyYAML class" % event.__class__.__name__)

    if event.__class__ in self._event_method_map:
      self._event_method_map[event.__class__](event, loader)

  def _HandleEvents(self, events):
    """Iterate over all events and send them to handler.

    This method is not meant to be called from the interface.

    Only use in tests.

    Args:
      events: Iterator or generator containing events to process.
    raises:
      EventListenerParserError when a yaml.parser.ParserError is raised.
      EventError when an exception occurs during the handling of an event.
    """
    for event in events:
      try:
        self.HandleEvent(*event)
      except Exception as e:
        event_object, loader = event
        raise yaml_errors.EventError(e, event_object)

  def _GenerateEventParameters(self,
                               stream,
                               loader_class=yaml.loader.SafeLoader,
                               **loader_args
                              ):
    """Creates a generator that yields event, loader parameter pairs.

    For use as parameters to HandleEvent method for use by Parse method.
    During testing, _GenerateEventParameters is simulated by allowing
    the harness to pass in a list of pairs as the parameter.

    A list of (event, loader) pairs must be passed to _HandleEvents otherwise
    it is not possible to pass the loader instance to the handler.

    Also responsible for instantiating the loader from the Loader
    parameter.

    Args:
      stream: String document or open file object to process as per the
        yaml.parse method.  Any object that implements a 'read()' method which
        returns a string document will work.
      loader_class: Loader class to use as per the yaml.parse method.  Used to
        instantiate new yaml.loader instance.
      **loader_args: Pass to the loader on construction


    Yields:
      Tuple(event, loader) where:
        event: Event emitted by PyYAML loader.
        loader: Used for dependency injection.
    """
    assert loader_class is not None
    try:
      loader = loader_class(stream, **loader_args)
      while loader.check_event():
        yield (loader.get_event(), loader)
    except yaml.error.YAMLError as e:
      raise yaml_errors.EventListenerYAMLError(e)

  def Parse(self, stream, loader_class=yaml.loader.SafeLoader, **loader_args):
    """Calls YAML parser to generate and handle all events.

    Calls PyYAML parser and sends resulting generator to `handle_event` method
    for processing.

    Args:
      stream: String document or open file object to process as per the
        `yaml.parse` method. Any object that implements a `read()` method which
        returns a string document will work with the YAML parser.
      loader_class: Used for dependency injection.
      **loader_args: Pass to the loader on construction.
    """


    version = (1, 1)
    if 'version' in loader_args:
      loader_args = copy.copy(loader_args)
      version = loader_args['version']
      del loader_args['version']
    self._HandleEvents(self._GenerateEventParameters(
        stream, loader_class, version=version, **loader_args))



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


_sym_db = _symbol_database.Default()


from google.appengine.api import api_base_pb2 as google_dot_appengine_dot_api_dot_api__base__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n;google/appengine/api/blobstore/blobstore_stub_service.proto\x12\x10google.appengine\x1a#google/appengine/api/api_base.proto\"5\n\x10StoreBlobRequest\x12\x10\n\x08\x62lob_key\x18\x01 \x02(\t\x12\x0f\n\x07\x63ontent\x18\x02 \x01(\x0c\"\x8f\x01\n\x19SetBlobStorageTypeRequest\x12M\n\x0cstorage_type\x18\x01 \x02(\x0e\x32\x37.google.appengine.SetBlobStorageTypeRequest.StorageType\"#\n\x0bStorageType\x12\n\n\x06MEMORY\x10\x00\x12\x08\n\x04\x46ILE\x10\x01\x42<\n\"com.google.appengine.api.blobstoreB\x16\x42lobstoreStubServicePb')



_STOREBLOBREQUEST = DESCRIPTOR.message_types_by_name['StoreBlobRequest']
_SETBLOBSTORAGETYPEREQUEST = DESCRIPTOR.message_types_by_name['SetBlobStorageTypeRequest']
_SETBLOBSTORAGETYPEREQUEST_STORAGETYPE = _SETBLOBSTORAGETYPEREQUEST.enum_types_by_name['StorageType']
StoreBlobRequest = _reflection.GeneratedProtocolMessageType('StoreBlobRequest', (_message.Message,), {
  'DESCRIPTOR' : _STOREBLOBREQUEST,
  '__module__' : 'google.appengine.api.blobstore.blobstore_stub_service_pb2'

  })
_sym_db.RegisterMessage(StoreBlobRequest)

SetBlobStorageTypeRequest = _reflection.GeneratedProtocolMessageType('SetBlobStorageTypeRequest', (_message.Message,), {
  'DESCRIPTOR' : _SETBLOBSTORAGETYPEREQUEST,
  '__module__' : 'google.appengine.api.blobstore.blobstore_stub_service_pb2'

  })
_sym_db.RegisterMessage(SetBlobStorageTypeRequest)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\"com.google.appengine.api.blobstoreB\026BlobstoreStubServicePb'
  _STOREBLOBREQUEST._serialized_start=118
  _STOREBLOBREQUEST._serialized_end=171
  _SETBLOBSTORAGETYPEREQUEST._serialized_start=174
  _SETBLOBSTORAGETYPEREQUEST._serialized_end=317
  _SETBLOBSTORAGETYPEREQUEST_STORAGETYPE._serialized_start=282
  _SETBLOBSTORAGETYPEREQUEST_STORAGETYPE._serialized_end=317




#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Get the App Engine app ID from environment.

Not to be confused with `google.appengine.api.app_identity.get_application_id()`
which gets a "display" app ID.
"""






import os
from typing import MutableMapping, Text, Optional


OsEnvironLike = Optional[MutableMapping[Text, Text]]


def get(environ: OsEnvironLike = None) -> str:
  """Get the application ID from the environment.

  Args:
    environ: Environment dictionary. Uses os.environ if `None`.

  Returns:
    Default application ID as a string.

  We read from the environment APPLICATION_ID (deprecated) or else
  GAE_APPLICATION.
  """

  if environ is None:
    environ = os.environ

  return environ.get('APPLICATION_ID', environ.get('GAE_APPLICATION', ''))


def put(app_id: str, environ: OsEnvironLike = None):
  """Set the application ID in the environment.

  Args:
    app_id: Application ID as a string.
    environ: Environment dictionary. Uses os.environ if `None`.
  """

  if environ is None:
    environ = os.environ

  environ['APPLICATION_ID'] = app_id
  environ['GAE_APPLICATION'] = app_id


def normalize(environ: OsEnvironLike = None):
  """Normalize the environment variables which set the app ID."""

  put(get(environ=environ), environ=environ)


def clear(environ: OsEnvironLike = None):
  """Unset the application ID in the environment.

  Args:
    environ: Environment dictionary. Uses os.environ if `None`.
  """

  if environ is None:
    environ = os.environ

  environ.pop('APPLICATION_ID', None)
  environ.pop('GAE_APPLICATION', None)


_PARTITION_SEPARATOR = '~'
_DOMAIN_SEPARATOR = ':'


def parse(app_id: Optional[str] = None,
          environ: OsEnvironLike = None):
  """Parses a full app ID into `partition`, `domain_name`, and `display_app_id`.

  Args:
    app_id: The full partitioned app ID. Looks up from environ if `None`.
    environ: Environment dictionary. Uses os.environ if `None`.

  Returns:
    A tuple `(partition, domain_name, display_app_id)`.  The partition and
    domain name might be empty.
  """
  if app_id is None:
    app_id = get(environ)
  partition = ''
  psep = app_id.find(_PARTITION_SEPARATOR)
  if psep > 0:
    partition = app_id[:psep]
    app_id = app_id[psep+1:]
  domain_name = ''
  dsep = app_id.find(_DOMAIN_SEPARATOR)
  if dsep > 0:
    domain_name = app_id[:dsep]
    app_id = app_id[dsep+1:]
  return partition, domain_name, app_id


def project_id(app_id: Optional[str] = None,
               environ: OsEnvironLike = None):
  """Parses the domain prefixed project ID from the app_id.

  Args:
    app_id: The full partitioned app ID. Looks up from environ if `None`.
    environ: Environment dictionary. Uses os.environ if `None`.

  Returns:
    A tuple `(partition, domain_name, display_app_id)`.  The partition and
    domain name might be empty.
  """
  _, domain_name, display_app_id = parse(app_id, environ)
  if domain_name:
    return f'{domain_name}{_DOMAIN_SEPARATOR}{display_app_id}'
  return display_app_id



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""In-memory implementation of Blobstore stub storage based on file system.

This module contains an implementation of `blob_storage.BlobStorage` that
writes blobs directly to a file system.
"""

import errno
import os

import six
import six.moves.builtins

from google.appengine.api import blobstore
from google.appengine.api.blobstore import blob_storage











__all__ = ['FileBlobStorage']



_local_open = six.moves.builtins.open


class FileBlobStorage(blob_storage.BlobStorage):
  """The storage mechanism that stores blob data on a local disk."""

  def __init__(self, storage_directory, app_id):
    """Constructor.

    Args:
      storage_directory: Directory within which to store blobs.
      app_id: The application ID on whose behalf to store blobs.
    """
    self._storage_directory = storage_directory
    self._app_id = app_id

  @classmethod
  def _BlobKey(cls, blob_key):
    """Normalizes to an instance of `BlobKey`."""
    if not isinstance(blob_key, blobstore.BlobKey):
      return blobstore.BlobKey(six.text_type(blob_key))
    return blob_key

  def _DirectoryForBlob(self, blob_key):
    """Determines the directory in which a blob is stored.

    Each blob is written to a directory underneath the storage objects
    storage directory based on the kind of blobs, application ID, and first
    character of its name. For example, the following blob keys...::

        _ACFDEDG
        _MNOPQRS
        _RSTUVWX


    ...are stored in the following locations::

        <storage-dir>/blob/myapp/A
        <storage-dir>/blob/myapp/M
        <storage-dir>/R


    Args:
      blob_key: Blob key for which you need to determine the directory.

    Returns:
      A directory where the blob is stored or should be stored, relative to this
      object's storage directory.
    """
    blob_key = self._BlobKey(blob_key)
    return os.path.join(self._storage_directory,
                        self._app_id,
                        str(blob_key)[1])

  def _FileForBlob(self, blob_key):
    """Calculates the full file name in which to store blob contents.

    This method does not check to see if the file already exists.

    Args:
      blob_key: The blob key of the blob for which to calculate the file name.

    Returns:
      A string that contains the complete path for the file that is used to
      store the blob.
    """
    blob_key = self._BlobKey(blob_key)
    return os.path.join(self._DirectoryForBlob(blob_key), str(blob_key)[1:])

  def StoreBlob(self, blob_key, blob_stream):
    """Stores a blob stream .

    Args:
      blob_key: The blob key of the blob that you want to store.
      blob_stream: A stream or a stream-like object that will generate blob
          content.
    """
    blob_key = self._BlobKey(blob_key)
    blob_directory = self._DirectoryForBlob(blob_key)
    if not os.path.exists(blob_directory):
      os.makedirs(blob_directory)
    blob_file = self._FileForBlob(blob_key)
    output = _local_open(blob_file, 'wb')

    try:


      while True:
        block = blob_stream.read(1 << 20)
        if not block:
          break
        output.write(six.ensure_binary(block))
    finally:
      output.close()

  def OpenBlob(self, blob_key):
    """Opens a blob file for streaming.

    Args:
      blob_key: The blob key of an existing blob that you want to open.

    Returns:
      An open file stream to read the blob.
    """
    return _local_open(self._FileForBlob(blob_key), 'rb')

  def DeleteBlob(self, blob_key):
    """Deletes blob data.

    Deleting an unknown blob will not raise an error.

    Args:
      blob_key: The blob key of an existing blob that you want to delete.
    """
    try:
      os.remove(self._FileForBlob(blob_key))
    except OSError as e:
      if e.errno != errno.ENOENT:
        raise e



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Contains some functions that come in handy with XML parsing."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals


def GetTag(node):
  """Strips namespace prefix."""
  return node.tag.rsplit('}', 1)[-1]


def GetChild(node, tag):
  """Returns first child of node with tag."""
  for child in node.getchildren():
    if GetTag(child) == tag:
      return child


def BooleanValue(node_text):
  """Returns boolean value."""
  return node_text.lower() in ('1', 'true')


def GetAttribute(node, attr):
  """Wrapper function to retrieve attributes from XML nodes."""
  return node.attrib.get(attr, '')


def GetChildNodeText(node, child_tag, default=''):
  """Finds child XML node with desired tag and returns its text."""
  for child in node.getchildren():
    if GetTag(child) == child_tag:
      return GetNodeText(child) or default
  return default


def GetNodeText(node):
  """Returns the node text after stripping whitespace."""

  return node.text.strip() if node.text else ''


def GetNodes(node, match_tag):
  """Gets all children of a node with the desired tag."""
  return (child for child in node.getchildren() if GetTag(child) == match_tag)



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""A mechanism for library configuration.

Whenever App Engine library code needs a user-configurable value, it should use
the following protocol:

1.  Pick a prefix unique to the library module, for example: `mylib`.
2.  Call `lib_config.register(prefix, mapping)` with that prefix as the first
    argument and a dict mapping suffixes to default functions as the second.
3.  The `register()` function returns a configuration handle that is unique
    to this prefix. The configuration handle object has attributes that
    correspond to each of the suffixes given in the mapping. Call these
    functions to access the user's configuration value. If the user didn't
    configure a function, the default function from the mapping is called
    instead.
4.  Document the function name and its signature and semantics.

Users that want to provide configuration values should create a module named
`appengine_config.py` in the top-level directory of their application and define
functions as documented by various App Engine library components in that module.
To change the configuration, edit the file and re-deploy the application. When
using the SDK, no redeployment is required; the development server will pick up
the changes the next time it handles a request.

Third party libraries can also use this mechanism. For casual use, calling the
`register()` method with a unique prefix is acceptable. For more complex
libraries, however, you should instantiate a new `LibConfigRegistry` instance
that uses a different module name.


Example `appengine_config.py` file:

    ```python
    from somewhere import MyMiddleWareClass

    def mylib_add_middleware(app):
      app = MyMiddleWareClass(app)
      return app
    ```

Example library use:

    ```python
    from google.appengine.api import lib_config

    config_handle = lib_config.register(
        'mylib',
        {'add_middleware': lambda app: app})

    def add_middleware(app): return config_handle.add_middleware(app)
    ```
"""

import importlib
import logging
import os
import sys
import threading



__all__ = ['DEFAULT_MODNAME',
           'LibConfigRegistry',
           'ConfigHandle',
           'register',
           'main',
           ]




DEFAULT_MODNAME = 'appengine_config'





class LibConfigRegistry(object):
  """A registry containing library configuration values.

  Note: The actual import of this module is deferred until the first
  time a configuration value is requested through attribute access
  on a `ConfigHandle` instance.
  """

  def __init__(self, modname):
    """Constructor.

    Args:
      modname: The module name that needs to be imported.
    """
    self._modname = modname
    self._registrations = {}
    self._module = None
    self._lock = threading.RLock()

  def register(self, prefix, mapping):
    """Registers a set of configuration names.

    Args:
      prefix: A shared prefix for the configuration names being registered.
          If the prefix doesn't end in `_`, that character is appended.
      mapping: A `dict` that maps suffix strings to default values.

    Returns:
      A `ConfigHandle` instance.

    You can re-register the same prefix. When you do so, the mappings are
    merged, and for duplicate suffixes, the most recent registration is used.
    """

    if not prefix.endswith('_'):
      prefix += '_'
    self._lock.acquire()
    try:
      handle = self._registrations.get(prefix)
      if handle is None:
        handle = ConfigHandle(prefix, self)
        self._registrations[prefix] = handle
    finally:
      self._lock.release()
    handle._update_defaults(mapping)
    return handle

  def initialize(self, import_func=importlib.import_module):
    """Tries to import the configuration module if it is not already imported.

    This function always sets `self._module` to a value that is not `None`;
    either the imported module (if it was imported successfully) or a
    placeholder `object()` instance (if an `ImportError` was raised) is used.
    Other exceptions are not caught.

    When a placeholder instance is used, the instance is also put in
    `sys.modules`. This usage allows us to detect when `sys.modules` was
    changed (as `dev_appserver.py` does when it notices source code changes)
    and retries the `import_module` in that case, while skipping it (for speed)
    if nothing has changed.

    Args:
      import_func: Used for dependency injection.
    """


    if not self._modname:
      return

    self._lock.acquire()

    try:
      if (self._module is not None and
          self._module is sys.modules.get(self._modname)):
        return
      try:
        import_func(self._modname)
      except ImportError as err:
        if str(err) not in [
            'No module named {}'.format(self._modname),
            "No module named '{}'".format(self._modname),
            'import of {} halted; None in sys.modules'.format(self._modname)]:

          raise
        logging.warning('Failed to import %s', self._modname)
        self._module = object()
        sys.modules[self._modname] = self._module
      else:
        self._module = sys.modules[self._modname]
    finally:
      self._lock.release()

  def reset(self):
    """Drops the imported configuration module.

    If the configuration module has not been imported, no operation occurs, and
    the next operation takes place.
    """
    self._lock.acquire()
    try:
      if self._module is None:

        return

      self._module = None
      handles = list(self._registrations.values())
    finally:
      self._lock.release()
    for handle in handles:
      handle._clear_cache()

  def _pairs(self, prefix):
    """Generates `(key, value)` pairs from the config module matching prefix.

    Args:
      prefix: A prefix string ending in `_`, for example: `mylib_`.

    Yields:
      `(key, value)` pairs, where `key` is the configuration name with the
      prefix removed, and `value` is the corresponding value.
    """
    self._lock.acquire()
    try:
      mapping = getattr(self._module, '__dict__', None)
      if not mapping:
        return
      items = list(mapping.items())
    finally:
      self._lock.release()
    nskip = len(prefix)
    for key, value in items:
      if key.startswith(prefix):
        yield key[nskip:], value

  def _dump(self):
    """Prints information about all registrations to stdout."""
    self.initialize()
    handles = []
    self._lock.acquire()
    try:
      if not hasattr(self._module, '__dict__'):
        print('Module %s.py does not exist.' % self._modname)
      elif not self._registrations:
        print('No registrations for %s.py.' % self._modname)
      else:
        print('Registrations in %s.py:' % self._modname)
        print('-'*40)
        handles = list(self._registrations.items())
    finally:
      self._lock.release()
    for _, handle in sorted(handles):
      handle._dump()


class ConfigHandle(object):
  """A set of configuration for a single library module or package.

  Public attributes of instances of this class are configuration values.
  Attributes are dynamically computed in `__getattr__()`, and cached as regular
  instance attributes.
  """

  _initialized = False

  def __init__(self, prefix, registry):
    """Constructor.

    Args:
      prefix: A shared prefix for the configuration names being registered. It
          must end in `_`. This requirement is enforced by
          `LibConfigRegistry`.
      registry: A `LibConfigRegistry` instance.
    """
    assert prefix.endswith('_')
    self._prefix = prefix
    self._defaults = {}
    self._overrides = {}
    self._registry = registry
    self._lock = threading.RLock()

  def _update_defaults(self, mapping):
    """Updates the default mappings.

    Args:
      mapping: A dict mapping suffix strings to default values.
    """
    self._lock.acquire()
    try:
      for key, value in mapping.items():
        if key.startswith('__') and key.endswith('__'):
          continue
        self._defaults[key] = value
      if self._initialized:
        self._update_configs()
    finally:
      self._lock.release()

  def _update_configs(self):
    """Updates the configuration values.

    This clears the cached values, initializes the registry, and loads
    the configuration values from the config module.
    """
    self._lock.acquire()
    try:
      if self._initialized:
        self._clear_cache()
      self._registry.initialize()
      for key, value in self._registry._pairs(self._prefix):
        if key not in self._defaults:
          logging.warn('Configuration "%s" not recognized', self._prefix + key)
        else:
          self._overrides[key] = value
      self._initialized = True
    finally:
      self._lock.release()

  def _clear_cache(self):
    """Clears the cached values."""
    self._lock.acquire()
    try:
      self._initialized = False
      for key in self._defaults:
        self._overrides.pop(key, None)
        try:
          delattr(self, key)
        except AttributeError:
          pass
    finally:
      self._lock.release()

  def _dump(self):
    """Prints information about this set of registrations to stdout."""
    self._lock.acquire()
    try:
      print('Prefix %s:' % self._prefix)
      if self._overrides:
        print('  Overrides:')
        for key in sorted(self._overrides):
          print('    %s = %r' % (key, self._overrides[key]))
      else:
        print('  No overrides')
      if self._defaults:
        print('  Defaults:')
        for key in sorted(self._defaults):
          print('    %s = %r' % (key, self._defaults[key]))
      else:
        print('  No defaults')
      print('-'*40)
    finally:
      self._lock.release()

  def __getattr__(self, suffix):
    """Dynamic attribute access.

    Args:
      suffix: The attribute name.

    Returns:
      A configuration values.

    Raises:
      AttributeError: If the suffix is not a registered suffix.

    The first time an attribute is referenced, this method is invoked. The value
    returned is taken either from the config module or from the registered
    default.
    """
    self._lock.acquire()
    try:
      if not self._initialized:
        self._update_configs()
      if suffix in self._overrides:
        value = self._overrides[suffix]
      elif suffix in self._defaults:
        value = self._defaults[suffix]
      else:
        raise AttributeError(suffix)

      setattr(self, suffix, value)
      return value
    finally:
      self._lock.release()



_default_registry = LibConfigRegistry(DEFAULT_MODNAME)


def register(prefix, mapping):
  """Registers a set of configurations with the default config module.

  Args:
    prefix: A shared prefix for the configuration names being registered.
        If the prefix doesn't end in `_`, that character is appended.
    mapping: A `dict` mapping suffix strings to default values.

  Returns:
    A `ConfigHandle` instance.
  """
  return _default_registry.register(prefix, mapping)


def main():
  """Dumps the configuration, using a CGI-style request handler.

  Put this in your `app.yaml` file to enable (you can pick any URL):

      ```python
      - url: /lib_config
        script: $PYTHON_LIB/google/appengine/api/lib_config.py
      ```

  Note:
      Unless you are using the SDK, you must be an administrator to use this
      function.
  """
  if not os.getenv('SERVER_SOFTWARE', '').startswith('Dev'):
    from google.appengine.api import users
    if not users.is_current_user_admin():
      if users.get_current_user() is None:
        print('Status: 302')
        print('Location:', users.create_login_url(os.getenv('PATH_INFO', '')))
      else:
        print('Status: 403')
        print()
        print('Forbidden')
      return

  print('Content-type: text/plain')
  print()
  _default_registry._dump()


if __name__ == '__main__':
  main()



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""OS cross-platform compatibility tweaks.

This module will, on import, change some parts of the running evironment so
that other modules do not need special handling when running on different
operating systems, such as Linux/Mac OSX/Windows.

Some of these changes must be done before other modules are imported, so
always import this module first.
"""








import os
import time

import six.moves.builtins

os.environ['TZ'] = 'UTC'
if hasattr(time, 'tzset'):
  time.tzset()







if 'WindowsError' in six.moves.builtins.__dict__:
  WindowsError = WindowsError
else:
  class WindowsError(Exception):
    """A fake Windows Error exception which should never be thrown."""




ERROR_PATH_NOT_FOUND = 3
ERROR_ACCESS_DENIED = 5
ERROR_SHARING_VIOLATION = 32
ERROR_ALREADY_EXISTS = 183
WSAECONNABORTED = 10053




#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Request info implementation for Titanoboa Applauncher."""

import threading

import requests
from requests import structures
import six

from google.appengine.api import request_info


class NotSupportedError(request_info.Error):
  pass


class LocalDispatcher(request_info.Dispatcher):
  """Dispatch requests across all instances and modules of application.

  As Titanoboa doesn't support multiple instances, this dispatcher always
  refers to a single module, version, and instance.
  """

  def __init__(self, default_hostname=None, app=None):
    self._default_hostname = default_hostname
    self._app = app

  def get_module_names(self):
    return ['default']

  def get_versions(self, module):
    return ['1']

  def get_default_version(self, module):
    return '1'

  def get_hostname(self, module, version, instance=None):
    return self._default_hostname

  def set_num_instances(self, module, version, instances):
    raise NotSupportedError('Cannot change number of instances.')

  def get_num_instances(self, module, version):
    return None

  def start_version(self, module, version):
    raise NotSupportedError('Multiple versions is not supported.')

  def stop_version(self, module, version):
    raise NotSupportedError('Multiple versions is not supported.')

  def add_event(self, runnable, eta, service=None, event_id=None):

    raise NotImplementedError()

  def update_event(self, eta, service, event_id):

    raise NotImplementedError()

  def add_request(
      self, method, relative_url, headers, body, source_ip,
      module_name=None, version=None, instance_id=None):
    requests_headers = structures.CaseInsensitiveDict()
    for k, v in headers:

      requests_headers[six.ensure_str(k)] = v
    requests_headers['X-AppEngine-User-IP'] = source_ip
    requests_headers['X-AppEngine-Fake-Is-Admin'] = '1'
    if 'host' in requests_headers:
      hostname = requests_headers['host']
    else:
      hostname = self.get_hostname(module_name, version)
    response = requests.request(
        method, 'http://' + hostname + six.ensure_str(relative_url),
        headers=requests_headers, data=body)


    return request_info.ResponseTuple(
        response.status_code, response.headers, response.content)

  def add_async_request(
      self, method, relative_url, headers, body, source_ip,
      module_name=None, version=None, instance_id=None):
    t = threading.Thread(target=self.add_request, kwargs=dict(
        method=method, relative_url=relative_url, headers=headers, body=body,
        source_ip=source_ip, module_name=module_name, version=version,
        instance_id=instance_id))
    t.start()

  def send_background_request(
      self, module_name, version, instance, background_request_id):
    if self._app:
      if not (self._app.app_yaml.basic_scaling or
              self._app.app_yaml.manual_scaling):
        raise request_info.NotSupportedWithAutoScalingError()
      if self._app.cfg.workers > 1:



        raise NotSupportedError(
            'Background requests not supported with multi-process server.')
    headers = (
        ('X-AppEngine-BackgroundRequest', background_request_id),
    )
    self.add_async_request(
        'get', '/_ah/background', headers, body='', source_ip='0.1.0.3',
        module_name=module_name, version=version, instance_id=instance)


class LocalRequestInfo(request_info._LocalRequestInfo):
  """Request info for Titanoboa local launcher."""

  def __init__(self, default_address=None, app=None):
    if default_address is None:
      default_address = '127.0.0.1:8080'
    self._address = 'http://' + default_address
    self._dispatcher = LocalDispatcher(default_address, app)

  def get_dispatcher(self):
    return self._dispatcher

  def get_address(self):
    return self._address



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Test utilties for YAML parsing tests.
"""



from google.appengine.api import yaml_listener
from google.appengine.api import yaml_builder

import google
from ruamel import yaml


class FakeDocument(object):
  """Simple document with single object root."""

  def __init__(self):
    """Initialize fake document.

    Fake document has single value, the root, which is where its contents
    is stored.
    """
    self.root = None


class FakeEventHandler(yaml_listener.EventHandler):
  """Fake event handler

  All this class does is receive events, check that they are the right type
  and append it to an event list so that they can be checked for proper call
  sequence later by unit tests.
  """

  def __init__(self, test, expect_loader=None):
    """Initializer for fake event handler.

    Args:
      test: Test from which fake handler is called.
      expect_loader: Which loader should be passed along as a parameter
       by listener calls.  Can be set to None for tests which do not
       involve use of a loader.
    """
    self.events = []
    self.test = test
    self.expect_loader = expect_loader

  def HandleEvent(self, event, loader, event_class):
    """Handle event

    If fake handler constructed with a loader, will check that it has received
    it as its loader parameter.  Checks that the received event is of the
    expected class.  If successful, appends event to list of received events.

    Args:
      event: Event received from emitter or test harness.
      loader: Loader received from emitter or test harness.
      event_class: Expected class which event should be instance of.
    """

    if self.expect_loader:
      self.test.assertEquals(self.expect_loader, loader)

    self.test.assertTrue(isinstance(event, event_class),
                         "Expected %s, was %s" %
                         (event.__class__.__name__,
                          event_class.__name__))
    self.events.append(event)

  def StreamStart(self, event, loader):
    """Handle start of stream event"""
    self.HandleEvent(event, loader, yaml.events.StreamStartEvent)

  def StreamEnd(self, event, loader):
    """Handle end of stream event"""
    self.HandleEvent(event, loader, yaml.events.StreamEndEvent)

  def DocumentStart(self, event, loader):
    """Handle start of document event"""
    self.HandleEvent(event, loader, yaml.events.DocumentStartEvent)

  def DocumentEnd(self, event, loader):
    """Handle end of document event"""
    self.HandleEvent(event, loader, yaml.events.DocumentEndEvent)

  def Alias(self, event, loader):
    """Handle alias event"""
    self.HandleEvent(event, loader, yaml.events.AliasEvent)

  def Scalar(self, event, loader):
    """Handle scalar event"""
    self.HandleEvent(event, loader, yaml.events.ScalarEvent)

  def SequenceStart(self, event, loader):
    """Handle start of sequence event"""
    self.HandleEvent(event, loader, yaml.events.SequenceStartEvent)

  def SequenceEnd(self, event, loader):
    """Handle end of sequence event"""
    self.HandleEvent(event, loader, yaml.events.SequenceEndEvent)

  def MappingStart(self, event, loader):
    """Handle start of mapping event"""
    self.HandleEvent(event, loader, yaml.events.MappingStartEvent)

  def MappingEnd(self, event, loader):
    """Handle end of mapping event"""
    self.HandleEvent(event, loader, yaml.events.MappingEndEvent)


class FakeBuilder(yaml_builder.Builder):
  """Maps YAML events directly to collection types for easy inspection.

  Note, this is very similar if not identical to what yaml.load creates.
  """

  def BuildDocument(self):
    """Construct new fake document"""
    return FakeDocument()

  def InitializeDocument(self, document, value):
    """Initialize root with scalar value"""
    document.root = value

  def BuildMapping(self, top_value):
    """Use dict as mapping object.

    Args:
      top_value: Not used

    Returns:
      An new dict object.
    """
    return {}

  def EndMapping(self, top_value, mapping):
    """Append 'end-mapping' to mapping.

    Args:
      top_value: Not used.
      mapping: Ending mapping.
    """
    mapping['end-mapping'] = True

  def BuildSequence(self, top_value):
    """Use list as sequence object.

    Args:
      top_value: Not used

    Returns:
      An new list object.
    """
    return []

  def EndSequence(self, top_value, sequence):
    """Append 'end-sequence' to sequence.

    Args:
      top_value: Not used.
      sequence: Ending sequence.
    """
    sequence.append('end-sequence')

  def MapTo(self, subject, key, value):
    """Map value to dict.

    Args:
      subject: dict object to map to.
      key: Key for dict object.
      value: Value to map to.
    """
    subject[key] = value

  def AppendTo(self, subject, value):
    """Append value to list.

    Args:
      subject: list object to append to.
      value: Value to append to list.
    """
    subject.append(value)



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#




"""An APIProxy stub that communicates with VMEngine service bridges."""

from concurrent import futures
import imp
import logging
import os
import sys
import threading
from google.appengine.api import apiproxy_rpc
from google.appengine.api import apiproxy_stub_map
from google.appengine.ext.remote_api import remote_api_bytes_pb2 as remote_api_pb2
from google.appengine.runtime import apiproxy_errors
import six.moves.urllib.parse
import urllib3









logging.getLogger('requests_nologs').setLevel(logging.ERROR)

TICKET_HEADER = 'HTTP_X_APPENGINE_API_TICKET'
DEV_TICKET_HEADER = 'HTTP_X_APPENGINE_DEV_REQUEST_ID'
DAPPER_ENV_KEY = 'HTTP_X_GOOGLE_DAPPERTRACEINFO'
SERVICE_BRIDGE_HOST = 'appengine.googleapis.internal'
API_PORT = 10001
SERVICE_ENDPOINT_NAME = 'app-engine-apis'
APIHOST_METHOD = '/VMRemoteAPI.CallRemoteAPI'
PROXY_PATH = '/rpc_http'
DAPPER_HEADER = 'X-Google-DapperTraceInfo'
SERVICE_DEADLINE_HEADER = 'X-Google-RPC-Service-Deadline'
SERVICE_ENDPOINT_HEADER = 'X-Google-RPC-Service-Endpoint'
SERVICE_METHOD_HEADER = 'X-Google-RPC-Service-Method'
RPC_CONTENT_TYPE = 'application/octet-stream'
DEFAULT_TIMEOUT = 60

DEADLINE_DELTA_SECONDS = 1





MAX_CONCURRENT_API_CALLS = 100

URLLIB3_POOL_COUNT = 10

URLLIB3_POOL_SIZE = 10



_EXCEPTIONS_MAP = {
    remote_api_pb2.RpcError.UNKNOWN:
        (apiproxy_errors.RPCFailedError,
         'The remote RPC to the application server failed for call %s.%s().'),
    remote_api_pb2.RpcError.CALL_NOT_FOUND:
        (apiproxy_errors.CallNotFoundError,
         'The API package \'%s\' or call \'%s()\' was not found.'),
    remote_api_pb2.RpcError.PARSE_ERROR:
        (apiproxy_errors.ArgumentError,
         'There was an error parsing arguments for API call %s.%s().'),
    remote_api_pb2.RpcError.OVER_QUOTA:
        (apiproxy_errors.OverQuotaError,
         'The API call %s.%s() required more quota than is available.'),
    remote_api_pb2.RpcError.REQUEST_TOO_LARGE:
        (apiproxy_errors.RequestTooLargeError,
         'The request to API call %s.%s() was too large.'),
    remote_api_pb2.RpcError.CAPABILITY_DISABLED:
        (apiproxy_errors.CapabilityDisabledError,
         'The API call %s.%s() is temporarily disabled.'),
    remote_api_pb2.RpcError.FEATURE_DISABLED:
        (apiproxy_errors.FeatureNotEnabledError,
         'The API call %s.%s() is currently not enabled.'),
    remote_api_pb2.RpcError.RESPONSE_TOO_LARGE:
        (apiproxy_errors.ResponseTooLargeError,
         'The response from API call %s.%s() was too large.'),
    remote_api_pb2.RpcError.CANCELLED:
        (apiproxy_errors.CancelledError,
         'The API call %s.%s() was explicitly cancelled.'),
    remote_api_pb2.RpcError.DEADLINE_EXCEEDED:
        (apiproxy_errors.DeadlineExceededError,
         'The API call %s.%s() took too long to respond and was cancelled.')
}

_DEFAULT_EXCEPTION = _EXCEPTIONS_MAP[remote_api_pb2.RpcError.UNKNOWN]

_DEADLINE_EXCEEDED_EXCEPTION = _EXCEPTIONS_MAP[
    remote_api_pb2.RpcError.DEADLINE_EXCEEDED]






class DefaultApiRPC(apiproxy_rpc.RPC):
  """A class representing an RPC to a remote server."""

  def _ErrorException(self, exception_class, error_details):
    return exception_class(error_details % (self.package, self.call))

  def _TranslateToError(self, response):
    """Translates a failed APIResponse into an exception."""


    if response.HasField('rpc_error'):
      code = response.rpc_error.code
      detail = response.rpc_error.detail
      exception_type, msg = _EXCEPTIONS_MAP.get(code, _DEFAULT_EXCEPTION)
      if detail:
        msg = '%s -- Additional details from server: %s' % (msg, detail)
      raise self._ErrorException(exception_type, msg)


    raise apiproxy_errors.ApplicationError(response.application_error.code,
                                           response.application_error.detail)

  def _MakeCallImpl(self):
    """Makes an asynchronous API call over the service bridge.

    For this to work the following must be set:
      self.package: the API package name;
      self.call: the name of the API call/method to invoke;
      self.request: the API request body as a serialized protocol buffer.

    The actual API call is made by urllib3.request via a thread pool
    (multiprocessing.dummy.Pool). The thread pool restricts the number of
    concurrent requests to MAX_CONCURRENT_API_CALLS, so this method will
    block if that limit is exceeded, until other asynchronous calls resolve.

    If the main thread holds the import lock, waiting on thread work can cause
    a deadlock:
    https://docs.python.org/2/library/threading.html#importing-in-threaded-code

    Therefore, we try to detect this error case and fall back to sync calls.
    """
    assert self._state == apiproxy_rpc.RPC.IDLE, self._state






    ticket = None
    if DefaultApiStub.ShouldUseRequestSecurityTicketForThread():


      ticket = os.environ.get(TICKET_HEADER, os.environ.get(DEV_TICKET_HEADER))

    if not ticket:
      raise apiproxy_errors.RPCFailedError(
          'Attempted RPC call without active security ticket')

    request = remote_api_pb2.Request(
        service_name=self.package,
        method=self.call,
        request_id=ticket,
        request=self.request.SerializeToString())

    deadline = self.deadline or DEFAULT_TIMEOUT

    body_data = request.SerializeToString()
    headers = {
        SERVICE_DEADLINE_HEADER: str(deadline),
        SERVICE_ENDPOINT_HEADER: SERVICE_ENDPOINT_NAME,
        SERVICE_METHOD_HEADER: APIHOST_METHOD,
        'Content-type': RPC_CONTENT_TYPE,
    }


    dapper_header_value = os.environ.get(DAPPER_ENV_KEY)
    if dapper_header_value:
      headers[DAPPER_HEADER] = dapper_header_value





    api_host = os.environ.get('API_HOST', SERVICE_BRIDGE_HOST)
    api_port = os.environ.get('API_PORT', API_PORT)

    if ':' in api_host:
      api_host = '[{}]'.format(api_host)
    endpoint_url = six.moves.urllib.parse.urlunparse(
        ('http', '%s:%s' % (api_host, api_port), PROXY_PATH, '', '', ''))

    self._state = apiproxy_rpc.RPC.RUNNING

    request_kwargs = dict(
        url=endpoint_url,
        method='POST',
        timeout=DEADLINE_DELTA_SECONDS + deadline,
        headers=headers,
        body=body_data)






    if six.PY2 and imp.lock_held():
      self.future = futures.Future()
      self.future.set_result(self._SendRequestAndFinish(**request_kwargs))

    else:


      self.future = self.stub.thread_pool.submit(self._SendRequestAndFinish,
                                                 **request_kwargs)

  def _WaitImpl(self):

    assert self.future is not None
    futures.wait([self.future])
    return True

  def _SendRequest(self, **kwargs):
    try:
      response = self.stub.http.request(**kwargs)

      if response.status != 200:
        raise apiproxy_errors.RPCFailedError(
            'Proxy returned HTTP status %s %s' %
            (response.status, response.reason))
    except urllib3.exceptions.TimeoutError:
      raise self._ErrorException(*_DEADLINE_EXCEEDED_EXCEPTION)
    except (urllib3.exceptions.RequestError,
            urllib3.exceptions.ConnectionError):

      raise self._ErrorException(*_DEFAULT_EXCEPTION)


    parsed_response = remote_api_pb2.Response.FromString(response.data)


    if (parsed_response.HasField('application_error') or
        parsed_response.HasField('rpc_error')):
      raise self._TranslateToError(parsed_response)


    self.response.ParseFromString(parsed_response.response)

  def _CaptureTrace(self, f, **kwargs):
    try:
      f(**kwargs)
    except Exception:


      _, exc, tb = sys.exc_info()
      self._exception = exc
      self._traceback = tb

  def _SendRequestAndFinish(self, **kwargs):
    try:
      self._CaptureTrace(self._SendRequest, **kwargs)
    finally:
      if self.callback:
        self._CaptureTrace(self.callback)
      self._state = apiproxy_rpc.RPC.FINISHING


class _UseRequestSecurityTicketLocal(threading.local):
  """Thread local holding if the default ticket should always be used."""

  def __init__(self):
    super(_UseRequestSecurityTicketLocal, self).__init__()
    self.use_ticket_header_value = False


class DefaultApiStub(object):
  """A stub for calling services through a VM service bridge.

  You can use this to stub out any service that the remote server supports.
  """


  _USE_REQUEST_SECURITY_TICKET_LOCAL = _UseRequestSecurityTicketLocal()

  @classmethod
  def SetUseRequestSecurityTicketForThread(cls, value):
    """Sets if the in environment security ticket should be used.

    Security tickets are set in the os.environ, which gets inherited by a
    child thread.  Child threads should not use the security ticket of their
    parent by default, because once the parent thread returns and the request
    is complete, the security ticket is no longer valid.

    Args:
      value: Boolean value describing if we should use the security ticket.
    """
    cls._USE_REQUEST_SECURITY_TICKET_LOCAL.use_ticket_header_value = value

  @classmethod
  def ShouldUseRequestSecurityTicketForThread(cls):
    """Gets if thie security ticket should be used for this thread."""
    return cls._USE_REQUEST_SECURITY_TICKET_LOCAL.use_ticket_header_value


  def __init__(self):
    self.thread_pool = futures.ThreadPoolExecutor(MAX_CONCURRENT_API_CALLS)
    self.http = urllib3.PoolManager(
        num_pools=URLLIB3_POOL_COUNT, maxsize=URLLIB3_POOL_SIZE)

  def MakeSyncCall(self, service, call, request, response):
    """Make a synchronous API call.

    Args:
      service: The name of the service you are trying to use.
      call: The name of the method.
      request: The request protocol buffer
      response: The response protocol buffer to be filled.
    """
    rpc = self.CreateRPC()
    rpc.MakeCall(service, call, request, response)
    rpc.Wait()
    rpc.CheckSuccess()

  def CreateRPC(self):
    """Create a new RPC object."""
    return DefaultApiRPC(stub=self)


def Register(stub):
  """Insert stubs so App Engine services are accessed via the service bridge."""
  apiproxy_stub_map.apiproxy.SetDefaultStub(stub)



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#



"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n+google/appengine/api/urlfetch_service.proto\x12\x10google.appengine\"\xc2\x02\n\x14URLFetchServiceError\"\xa9\x02\n\tErrorCode\x12\x06\n\x02OK\x10\x00\x12\x0f\n\x0bINVALID_URL\x10\x01\x12\x0f\n\x0b\x46\x45TCH_ERROR\x10\x02\x12\x15\n\x11UNSPECIFIED_ERROR\x10\x03\x12\x16\n\x12RESPONSE_TOO_LARGE\x10\x04\x12\x15\n\x11\x44\x45\x41\x44LINE_EXCEEDED\x10\x05\x12\x19\n\x15SSL_CERTIFICATE_ERROR\x10\x06\x12\r\n\tDNS_ERROR\x10\x07\x12\n\n\x06\x43LOSED\x10\x08\x12\x1c\n\x18INTERNAL_TRANSIENT_ERROR\x10\t\x12\x16\n\x12TOO_MANY_REDIRECTS\x10\n\x12\x13\n\x0fMALFORMED_REPLY\x10\x0b\x12\x14\n\x10\x43ONNECTION_ERROR\x10\x0c\x12\x15\n\x11PAYLOAD_TOO_LARGE\x10\r\"\x80\x03\n\x0fURLFetchRequest\x12?\n\x06Method\x18\x01 \x02(\x0e\x32/.google.appengine.URLFetchRequest.RequestMethod\x12\x0b\n\x03Url\x18\x02 \x02(\t\x12\x38\n\x06header\x18\x03 \x03(\n2(.google.appengine.URLFetchRequest.Header\x12\x13\n\x07Payload\x18\x06 \x01(\x0c\x42\x02\x08\x01\x12\x1d\n\x0f\x46ollowRedirects\x18\x07 \x01(\x08:\x04true\x12\x10\n\x08\x44\x65\x61\x64line\x18\x08 \x01(\x01\x12+\n\x1dMustValidateServerCertificate\x18\t \x01(\x08:\x04true\x1a$\n\x06Header\x12\x0b\n\x03Key\x18\x04 \x02(\t\x12\r\n\x05Value\x18\x05 \x02(\t\"L\n\rRequestMethod\x12\x07\n\x03GET\x10\x01\x12\x08\n\x04POST\x10\x02\x12\x08\n\x04HEAD\x10\x03\x12\x07\n\x03PUT\x10\x04\x12\n\n\x06\x44\x45LETE\x10\x05\x12\t\n\x05PATCH\x10\x06\"\xdd\x02\n\x10URLFetchResponse\x12\x0f\n\x07\x43ontent\x18\x01 \x01(\x0c\x12\x12\n\nStatusCode\x18\x02 \x02(\x05\x12\x39\n\x06header\x18\x03 \x03(\n2).google.appengine.URLFetchResponse.Header\x12\"\n\x13\x43ontentWasTruncated\x18\x06 \x01(\x08:\x05\x66\x61lse\x12\x19\n\x11\x45xternalBytesSent\x18\x07 \x01(\x03\x12\x1d\n\x15\x45xternalBytesReceived\x18\x08 \x01(\x03\x12\x10\n\x08\x46inalUrl\x18\t \x01(\t\x12\x1d\n\x12\x41piCpuMilliseconds\x18\n \x01(\x03:\x01\x30\x12\x17\n\x0c\x41piBytesSent\x18\x0b \x01(\x03:\x01\x30\x12\x1b\n\x10\x41piBytesReceived\x18\x0c \x01(\x03:\x01\x30\x1a$\n\x06Header\x12\x0b\n\x03Key\x18\x04 \x02(\t\x12\r\n\x05Value\x18\x05 \x02(\tB9\n!com.google.appengine.api.urlfetchB\x11URLFetchServicePb\x88\x01\x01')



_URLFETCHSERVICEERROR = DESCRIPTOR.message_types_by_name['URLFetchServiceError']
_URLFETCHREQUEST = DESCRIPTOR.message_types_by_name['URLFetchRequest']
_URLFETCHREQUEST_HEADER = _URLFETCHREQUEST.nested_types_by_name['Header']
_URLFETCHRESPONSE = DESCRIPTOR.message_types_by_name['URLFetchResponse']
_URLFETCHRESPONSE_HEADER = _URLFETCHRESPONSE.nested_types_by_name['Header']
_URLFETCHSERVICEERROR_ERRORCODE = _URLFETCHSERVICEERROR.enum_types_by_name['ErrorCode']
_URLFETCHREQUEST_REQUESTMETHOD = _URLFETCHREQUEST.enum_types_by_name['RequestMethod']
URLFetchServiceError = _reflection.GeneratedProtocolMessageType('URLFetchServiceError', (_message.Message,), {
  'DESCRIPTOR' : _URLFETCHSERVICEERROR,
  '__module__' : 'google.appengine.api.urlfetch_service_pb2'

  })
_sym_db.RegisterMessage(URLFetchServiceError)

URLFetchRequest = _reflection.GeneratedProtocolMessageType('URLFetchRequest', (_message.Message,), {

  'Header' : _reflection.GeneratedProtocolMessageType('Header', (_message.Message,), {
    'DESCRIPTOR' : _URLFETCHREQUEST_HEADER,
    '__module__' : 'google.appengine.api.urlfetch_service_pb2'

    })
  ,
  'DESCRIPTOR' : _URLFETCHREQUEST,
  '__module__' : 'google.appengine.api.urlfetch_service_pb2'

  })
_sym_db.RegisterMessage(URLFetchRequest)
_sym_db.RegisterMessage(URLFetchRequest.Header)

URLFetchResponse = _reflection.GeneratedProtocolMessageType('URLFetchResponse', (_message.Message,), {

  'Header' : _reflection.GeneratedProtocolMessageType('Header', (_message.Message,), {
    'DESCRIPTOR' : _URLFETCHRESPONSE_HEADER,
    '__module__' : 'google.appengine.api.urlfetch_service_pb2'

    })
  ,
  'DESCRIPTOR' : _URLFETCHRESPONSE,
  '__module__' : 'google.appengine.api.urlfetch_service_pb2'

  })
_sym_db.RegisterMessage(URLFetchResponse)
_sym_db.RegisterMessage(URLFetchResponse.Header)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n!com.google.appengine.api.urlfetchB\021URLFetchServicePb\210\001\001'
  _URLFETCHREQUEST.fields_by_name['Payload']._options = None
  _URLFETCHREQUEST.fields_by_name['Payload']._serialized_options = b'\010\001'
  _URLFETCHSERVICEERROR._serialized_start=66
  _URLFETCHSERVICEERROR._serialized_end=388
  _URLFETCHSERVICEERROR_ERRORCODE._serialized_start=91
  _URLFETCHSERVICEERROR_ERRORCODE._serialized_end=388
  _URLFETCHREQUEST._serialized_start=391
  _URLFETCHREQUEST._serialized_end=775
  _URLFETCHREQUEST_HEADER._serialized_start=661
  _URLFETCHREQUEST_HEADER._serialized_end=697
  _URLFETCHREQUEST_REQUESTMETHOD._serialized_start=699
  _URLFETCHREQUEST_REQUESTMETHOD._serialized_end=775
  _URLFETCHRESPONSE._serialized_start=778
  _URLFETCHRESPONSE._serialized_end=1127
  _URLFETCHRESPONSE_HEADER._serialized_start=661
  _URLFETCHRESPONSE_HEADER._serialized_end=697




#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""Define the DeadlineExceededError exception."""





try:
  BaseException
except NameError:
  BaseException = Exception


class DeadlineExceededError(BaseException):
  """Exception raised when the request reaches its overall time limit.

  This exception will be thrown by the original thread handling the request,
  shortly after the request reaches its deadline. Since the exception is
  asynchronously set on the thread by the App Engine runtime, it can appear
  to originate from any line of code that happens to be executing at that
  time.

  If the application catches this exception and does not generate a response
  very quickly afterwards, an error will be returned to the user and
  the application instance may be terminated.

  Not to be confused with runtime.apiproxy_errors.DeadlineExceededError.
  That one is raised when individual API calls take too long.
  """

  def __str__(self):
    return ('The overall deadline for responding to the HTTP request '
            'was exceeded.')



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""App identity stub service implementation.

This service behaves the same as the production service, except using
application default credentials.
"""

import datetime
import json
import threading
import time
import typing

from google.appengine.api import urlfetch
from google.appengine.api.app_identity import app_identity_service_pb2
from google.appengine.api.app_identity import app_identity_stub_base
from google.appengine.runtime import apiproxy_errors

from google import auth as google_auth
from google.auth import app_engine
from google.auth import credentials as ga_credentials
from google.auth import exceptions
from google.auth.transport import requests as transport
from google.oauth2 import service_account as oauth2_service_account

from pyasn1.codec.der import decoder
from pyasn1_modules.rfc2459 import Certificate
import rsa
import six
from six.moves import range
from six.moves import urllib


def BitStringToByteString(bs):
  """Convert a pyasn1.type.univ.BitString object to a string of bytes."""
  def BitsToInt(bits):
    return sum(v * (2 ** (7 - j)) for j, v in enumerate(bits))
  return bytes(
      bytearray([BitsToInt(bs[i:i + 8]) for i in range(0, len(bs), 8)]))


class DefaultCredentialsBasedAppIdentityServiceStub(
    app_identity_stub_base.AppIdentityServiceStubBase):
  """A stub for the AppIdentityService API for offline development.

  Provides stub functions which allow a developer to test integration before
  deployment.
  """

  THREADSAFE = True

  _credentials = None

  def __init__(self, service_name='app_identity_service'):
    super(DefaultCredentialsBasedAppIdentityServiceStub,
          self).__init__(service_name)
    self._credentials, _ = google_auth.default()
    if isinstance(self._credentials, app_engine.Credentials):



      raise exceptions.DefaultCredentialsError(
          'Could not automatically determine credentials.')
    self._access_token_cache_lock = threading.Lock()
    self._access_token_cache = {}
    self._x509_init_lock = threading.Lock()
    self._default_gcs_bucket_name = (
        app_identity_stub_base.APP_DEFAULT_GCS_BUCKET_NAME)
    self._x509 = None
    self._signing_key = None
    self._non_service_account_credentials = not isinstance(
        self._credentials, oauth2_service_account.Credentials)

  def _PopulateX509(self):
    with self._x509_init_lock:
      if self._x509 is None:

        url = ('https://www.googleapis.com/service_accounts/v1/metadata/x509/%s'
               % urllib.parse.unquote_plus(
                   self._credentials.service_account_email))
        response = urlfetch.fetch(
            url=url,
            validate_certificate=True,
            method=urlfetch.GET)
        if response.status_code != 200:
          raise apiproxy_errors.ApplicationError(
              app_identity_service_pb2.AppIdentityServiceError.UNKNOWN_ERROR,
              'Unable to load X509 cert: %s Response code: %i, Content: %s' % (
                  url, response.status_code, response.content))

        message = b'dummy'
        signature = self._credentials.sign_bytes(message)

        for signing_key, x509 in json.loads(response.content).items():
          der = rsa.pem.load_pem(x509, 'CERTIFICATE')
          asn1_cert, _ = decoder.decode(der, asn1Spec=Certificate())

          key_bitstring = (
              asn1_cert['tbsCertificate']
              ['subjectPublicKeyInfo']
              ['subjectPublicKey'])
          key_bytearray = BitStringToByteString(key_bitstring)

          public_key = rsa.PublicKey.load_pkcs1(key_bytearray, 'DER')
          try:
            if rsa.pkcs1.verify(message, signature, public_key):
              self._x509 = x509
              self._signing_key = signing_key
              return
          except rsa.pkcs1.VerificationError:
            pass

        raise apiproxy_errors.ApplicationError(
            app_identity_service_pb2.AppIdentityServiceError.UNKNOWN_ERROR,
            'Unable to find matching X509 cert for private key: %s' % url)

  def _Dynamic_SignForApp(self, request, response):
    """Implementation of AppIdentityService::SignForApp."""
    if self._non_service_account_credentials:




      return super(DefaultCredentialsBasedAppIdentityServiceStub,
                   self)._Dynamic_SignForApp(request, response)
    self._PopulateX509()
    signature = self._credentials.sign_bytes(
        request.bytes_to_sign)
    private_key_id = self._credentials.signer.key_id
    assert private_key_id == self._signing_key
    response.signature_bytes = signature
    response.key_name = self._signing_key

  def _Dynamic_GetPublicCertificatesForApp(self, request, response):
    """Implementation of AppIdentityService::GetPublicCertificatesForApp."""
    if self._non_service_account_credentials:




      return super(DefaultCredentialsBasedAppIdentityServiceStub,
                   self)._Dynamic_GetPublicCertificatesForApp(request, response)
    self._PopulateX509()
    certificate = response.public_certificate_list.add()
    certificate.key_name = self._signing_key
    certificate.x509_certificate_pem = self._x509

  def _Dynamic_GetServiceAccountName(self, request, response):
    """Implementation of AppIdentityService::GetServiceAccountName."""
    response.service_account_name = self.get_service_account_name()

  def _Dynamic_GetDefaultGcsBucketName(self, unused_request, response):
    """Implementation of AppIdentityService::GetDefaultGcsBucketName."""
    response.default_gcs_bucket_name = self._default_gcs_bucket_name

  def SetDefaultGcsBucketName(self, default_gcs_bucket_name):
    if default_gcs_bucket_name:
      self._default_gcs_bucket_name = default_gcs_bucket_name
    else:
      self._default_gcs_bucket_name = (
          app_identity_stub_base.APP_DEFAULT_GCS_BUCKET_NAME)

  def get_service_account_token(self, scopes, service_account=None):
    """Implementation of AppIdentityService::GetAccessToken.

    This API requires internet access.

    Args:
      scopes: a list of oauth2 scopes.
      service_account: the service account to get the token for

    Returns:
      Tuple of access token and expiration time in epoch

    Raises:
      RuntimeError: If unexpected response from Google server.
    """
    scope = ' '.join(scopes)
    with self._access_token_cache_lock:
      rv = self._access_token_cache.get(scope, None)

    if not (rv and rv['expires'] > (time.time() + 60)):
      credentials = self._credentials
      if (isinstance(credentials, ga_credentials.Scoped) and
          credentials.requires_scopes):
        credentials = credentials.with_scopes(scopes)
      if not credentials.valid:
        credentials.refresh(transport.Request())

      expiration = 0
      if credentials.expiry:
        expiration = TimestampFromNaiveUtcDatetime(credentials.expiry)
      rv = {
          'access_token': credentials.token,
          'expires': expiration,
      }
      with self._access_token_cache_lock:
        self._access_token_cache[scope] = rv

    return rv['access_token'], rv['expires']

  def get_service_account_name(self):
    if self._non_service_account_credentials:


      return ''
    return self._credentials.service_account_email


def TimestampFromNaiveUtcDatetime(dt):
  if dt.tzinfo is not None:
    raise ValueError('Expected a naive UTC datetime')
  if six.PY2:
    return (dt - datetime.datetime(1970, 1, 1)).total_seconds()

  return dt.replace(tzinfo=datetime.timezone.utc).timestamp()



#!/usr/bin/env python
#
# Copyright 2007 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


"""
A library for working with `BackendInfoExternal` records, describing backends
configured for an application. Supports loading the records from `backend.yaml`.
"""


from google.appengine.api import validation
from google.appengine.api import yaml_builder
from google.appengine.api import yaml_listener
from google.appengine.api import yaml_object









NAME_REGEX = r'(?!-)[a-z\d\-]{1,100}'
FILE_REGEX = r'(?!\^).*(?!\$).{1,256}'
CLASS_REGEX = r'^[bB](1|2|4|8|4_1G)$'
OPTIONS_REGEX = r'^[a-z, ]*$'
STATE_REGEX = r'^(START|STOP|DISABLED)$'


BACKENDS = 'backends'


NAME = 'name'
CLASS = 'class'
INSTANCES = 'instances'
OPTIONS = 'options'
PUBLIC = 'public'
DYNAMIC = 'dynamic'
FAILFAST = 'failfast'
MAX_CONCURRENT_REQUESTS = 'max_concurrent_requests'
START = 'start'

VALID_OPTIONS = frozenset([PUBLIC, DYNAMIC, FAILFAST])


STATE = 'state'


class BadConfig(Exception):
  """An invalid configuration was provided."""


class BackendEntry(validation.Validated):
  """A backend entry describes a single backend."""
  ATTRIBUTES = {
      NAME: NAME_REGEX,
      CLASS: validation.Optional(CLASS_REGEX),
      INSTANCES: validation.Optional(validation.TYPE_INT),
      MAX_CONCURRENT_REQUESTS: validation.Optional(validation.TYPE_INT),


      OPTIONS: validation.Optional(OPTIONS_REGEX),
      PUBLIC: validation.Optional(validation.TYPE_BOOL),
      DYNAMIC: validation.Optional(validation.TYPE_BOOL),
      FAILFAST: validation.Optional(validation.TYPE_BOOL),
      START: validation.Optional(FILE_REGEX),


      STATE: validation.Optional(STATE_REGEX),
  }

  def __init__(self, *args, **kwargs):
    super(BackendEntry, self).__init__(*args, **kwargs)
    self.Init()

  def Init(self):
    if self.public:
      raise BadConfig("Illegal field: 'public'")
    if self.dynamic:
      raise BadConfig("Illegal field: 'dynamic'")
    if self.failfast:
      raise BadConfig("Illegal field: 'failfast'")
    self.ParseOptions()
    return self

  def set_class(self, Class):
    """Setter for `Class`, since an attribute reference is an error."""
    self.Set(CLASS, Class)

  def get_class(self):
    """Accessor for `Class`, since an attribute reference is an error."""
    return self.Get(CLASS)

  def ToDict(self):
    """Returns a sorted dictionary representing the backend entry."""
    self.ParseOptions().WriteOptions()
    result = super(BackendEntry, self).ToDict()
    return validation.SortedDict([NAME,
                                  CLASS,
                                  INSTANCES,
                                  START,
                                  OPTIONS,
                                  MAX_CONCURRENT_REQUESTS,
                                  STATE],
                                 result)

  def ParseOptions(self):
    """Parses the `options` field and sets appropriate fields."""
    if self.options:
      options = [option.strip() for option in self.options.split(',')]
    else:
      options = []

    for option in options:
      if option not in VALID_OPTIONS:
        raise BadConfig('Unrecognized option: %s' % option)

    self.public = PUBLIC in options
    self.dynamic = DYNAMIC in options
    self.failfast = FAILFAST in options
    return self

  def WriteOptions(self):
    """Writes the `options` field based on other settings."""
    options = []
    if self.public:
      options.append('public')
    if self.dynamic:
      options.append('dynamic')
    if self.failfast:
      options.append('failfast')
    if options:
      self.options = ', '.join(options)
    else:
      self.options = None
    return self


def LoadBackendEntry(backend_entry):
  """Parses a `BackendEntry` object from a string.

  Args:
    backend_entry: a backend entry, as a string

  Returns:
    A `BackendEntry` object.
  """
  builder = yaml_object.ObjectBuilder(BackendEntry)
  handler = yaml_builder.BuilderHandler(builder)
  listener = yaml_listener.EventListener(handler)
  listener.Parse(backend_entry)

  entries = handler.GetResults()
  if len(entries) < 1:
    raise BadConfig('Empty backend configuration.')
  if len(entries) > 1:
    raise BadConfig('Multiple backend entries were found in configuration.')

  return entries[0].Init()


class BackendInfoExternal(validation.Validated):
  """`BackendInfoExternal` describes all backend entries for an application."""
  ATTRIBUTES = {
      BACKENDS: validation.Optional(validation.Repeated(BackendEntry)),
  }


def LoadBackendInfo(backend_info, open_fn=None):
  """Parses a `BackendInfoExternal` object from a string.

  Args:
    backend_info: a backends stanza (list of backends) as a string
    open_fn: Function for opening files. Unused.

  Returns:
    A `BackendInfoExternal` object.
  """
  builder = yaml_object.ObjectBuilder(BackendInfoExternal)
  handler = yaml_builder.BuilderHandler(builder)
  listener = yaml_listener.EventListener(handler)
  listener.Parse(backend_info)

  backend_info = handler.GetResults()
  if len(backend_info) < 1:
    return BackendInfoExternal(backends=[])
  if len(backend_info) > 1:
    raise BadConfig("Only one 'backends' clause is allowed.")

  info = backend_info[0]
  if not info.backends:
    return BackendInfoExternal(backends=[])

  for backend in info.backends:
    backend.Init()
  return info
